#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import string
from unittest import mock

import netaddr

from neutron_lib._i18n import _
from neutron_lib.api import converters
from neutron_lib.api.definitions import extra_dhcp_opt
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from neutron_lib import fixture
from neutron_lib.plugins import directory
from neutron_lib.tests import _base as base


def dummy_validator(data, valid_values=None):
    pass


class TestAttributeValidation(base.BaseTestCase):

    def _construct_dict_and_constraints(self):
        """Constructs a test dictionary and a definition of constraints.

        :return: A (dictionary, constraint) tuple
        """

        constraints = {'key1': {'type:values': ['val1', 'val2'],
                                'required': True},
                       'key2': {'type:string': None,
                                'required': False},
                       'key3': {'type:dict': {'k4': {'type:string': None,
                                                     'required': True}},
                                'required': True}}

        dictionary = {'key1': 'val1',
                      'key2': 'a string value',
                      'key3': {'k4': 'a string value'}}

        return dictionary, constraints

    def test_type_prefixing(self):
        validators.add_validator('type:prefixed_type', dummy_validator)
        validators.add_validator('unprefixed_type', dummy_validator)
        self.assertEqual(dummy_validator,
                         validators.get_validator('type:prefixed_type'))
        self.assertEqual(dummy_validator,
                         validators.get_validator('prefixed_type'))
        self.assertEqual(dummy_validator,
                         validators.get_validator('type:unprefixed_type'))
        self.assertEqual(dummy_validator,
                         validators.get_validator('unprefixed_type'))

    def test_adding_validator(self):
        validators.add_validator('new_type', dummy_validator)
        self.assertIn('type:new_type', validators.validators)
        self.assertEqual(dummy_validator,
                         validators.validators['type:new_type'])

    def test_get_validator_default(self):
        self.assertEqual(dummy_validator,
                         validators.get_validator('nope',
                                                  default=dummy_validator))

    def test_fail_adding_duplicate_validator(self):
        self.assertRaises(KeyError,
                          validators.add_validator,
                          'dict', lambda x: x)

    def test_success_adding_duplicate_validator(self):
        validators.add_validator('dummy', dummy_validator)
        validators.add_validator('dummy', dummy_validator)
        self.assertEqual(dummy_validator, validators.get_validator('dummy'))

    def test_is_attr_set(self):
        data = constants.ATTR_NOT_SPECIFIED
        self.assertIs(validators.is_attr_set(data), False)

        data = None
        self.assertIs(validators.is_attr_set(data), False)

        data = "I'm set"
        self.assertIs(validators.is_attr_set(data), True)

    def test_validate_values(self):
        # Check that validation is not performed if valid_values is not set
        msg = validators.validate_values(4)
        self.assertIsNone(msg)

        # Check that value is within valid_values
        msg = validators.validate_values(4, [4, 6])
        self.assertIsNone(msg)

        # Check that value is within valid_values
        msg = validators.validate_values(4, (4, 6))
        self.assertIsNone(msg)

        # Check that value is within valid_values with strings
        msg = validators.validate_values("1", ["2", "1", "4", "5"])
        self.assertIsNone(msg)

        # Check that value is not compatible for comparison
        response = "'valid_values' does not support membership operations"
        self.assertRaisesRegex(TypeError, response,
                               validators.validate_values, data=None,
                               valid_values=True)

    def test_validate_values_display(self):
        # Check that value is NOT within valid_values and report values
        msg = validators.validate_values(7, [4, 6],
                                         valid_values_display="[4, 6]")
        self.assertEqual("7 is not in [4, 6]", msg)

        # Check that value is NOT within valid_values and report values
        msg = validators.validate_values(7, (4, 6),
                                         valid_values_display="(4, 6)")
        self.assertEqual("7 is not in (4, 6)", msg)

        # Check values with a range function showing a custom string
        msg = validators.validate_values(8, range(8),
                                         valid_values_display="[0..7]")
        self.assertEqual("8 is not in [0..7]", msg)

        # Check that value is not within valid_values and custom string
        msg = validators.validate_values(1, [2, 3, 4, 5],
                                         valid_values_display="[2, 3, 4, 5]")
        self.assertEqual("1 is not in [2, 3, 4, 5]", msg)

        # Check that value is not within valid_values and custom string
        msg = validators.validate_values("1", ["2", "3", "4", "5"],
                                         valid_values_display="'valid_values"
                                         "_to_show'")
        self.assertEqual("1 is not in 'valid_values_to_show'", msg)

        # Check that value is not comparable to valid_values and got Exception
        data = 1
        valid_values = '[2, 3, 4, 5]'
        response = "'data' of type '%s' and 'valid_values' of type" \
                   " '%s' are not compatible for comparison" % (
                       type(data), type(valid_values))
        self.assertRaisesRegex(TypeError, response,
                               validators.validate_values, data,
                               valid_values,
                               valid_values_display="[2, 3, 4, 5]")

    def test_validate_not_empty_string(self):
        msg = validators.validate_not_empty_string('    ', None)
        self.assertEqual("'    ' Blank strings are not permitted", msg)
        msg = validators.validate_not_empty_string(123, None)
        self.assertEqual("'123' is not a valid string", msg)

    def test_validate_not_empty_string_or_none(self):
        msg = validators.validate_not_empty_string_or_none('    ', None)
        self.assertEqual("'    ' Blank strings are not permitted", msg)

        msg = validators.validate_not_empty_string_or_none(None, None)
        self.assertIsNone(msg)

    def test_validate_string_or_none(self):
        msg = validators.validate_string_or_none('test', None)
        self.assertIsNone(msg)

        msg = validators.validate_string_or_none(None, None)
        self.assertIsNone(msg)

    def test_validate_string(self):
        msg = validators.validate_string(None, None)
        self.assertEqual("'None' is not a valid string", msg)

        # 0 == len(data) == max_len
        msg = validators.validate_string("", 0)
        self.assertIsNone(msg)

        # 0 == len(data) < max_len
        msg = validators.validate_string("", 9)
        self.assertIsNone(msg)

        # 0 < len(data) < max_len
        msg = validators.validate_string("123456789", 10)
        self.assertIsNone(msg)

        # 0 < len(data) == max_len
        msg = validators.validate_string("123456789", 9)
        self.assertIsNone(msg)

        # 0 < max_len < len(data)
        msg = validators.validate_string("1234567890", 9)
        self.assertEqual("'1234567890' exceeds maximum length of 9", msg)

        msg = validators.validate_string("123456789", None)
        self.assertIsNone(msg)

    def test_validate_list_of_unique_strings(self):
        data = "TEST"
        msg = validators.validate_list_of_unique_strings(data, None)
        self.assertEqual("'TEST' is not a list", msg)

        data = ["TEST01", "TEST02", "TEST01"]
        msg = validators.validate_list_of_unique_strings(data, None)
        self.assertEqual(
            "Duplicate items in the list: 'TEST01'", msg)

        data = ["12345678", "123456789"]
        msg = validators.validate_list_of_unique_strings(data, 8)
        self.assertEqual("'123456789' exceeds maximum length of 8", msg)

        data = ["TEST01", "TEST02", "TEST03"]
        msg = validators.validate_list_of_unique_strings(data, None)
        self.assertIsNone(msg)

    def test_validate_oneline_not_empty_string(self):
        data = "Test"
        msg = validators.validate_oneline_not_empty_string(data, None)
        self.assertIsNone(msg)

        data = "Test but this is too long"
        max_len = 4
        msg = validators.validate_oneline_not_empty_string(data, max_len)
        self.assertEqual(
            "'{}' exceeds maximum length of {}".format(data, max_len),
            msg)

        data = "First line\nsecond line"
        msg = validators.validate_oneline_not_empty_string(data, None)
        self.assertEqual(
            "Multi-line string is not allowed: '%s'" % data,
            msg)

        data = ""
        msg = validators.validate_oneline_not_empty_string(data, None)
        self.assertEqual(
            "'%s' Blank strings are not permitted" % data,
            msg)

    def test_validate_oneline_not_empty_string_or_none(self):
        data = "Test"
        msg = validators.validate_oneline_not_empty_string_or_none(data, None)
        self.assertIsNone(msg)

        data = None
        msg = validators.validate_oneline_not_empty_string_or_none(data, None)
        self.assertIsNone(msg)

        data = "Test but this is too long"
        max_len = 4
        msg = validators.validate_oneline_not_empty_string_or_none(
            data, max_len)
        self.assertEqual(
            "'{}' exceeds maximum length of {}".format(data, max_len),
            msg)

        data = "First line\nsecond line"
        msg = validators.validate_oneline_not_empty_string_or_none(data, None)
        self.assertEqual(
            "Multi-line string is not allowed: '%s'" % data,
            msg)

        data = ""
        msg = validators.validate_oneline_not_empty_string(data, None)
        self.assertEqual(
            "'%s' Blank strings are not permitted" % data,
            msg)

    def test_validate_boolean(self):
        msg = validators.validate_boolean(True)
        self.assertIsNone(msg)
        msg = validators.validate_boolean(0)
        self.assertIsNone(msg)
        msg = validators.validate_boolean("false")
        self.assertIsNone(msg)
        msg = validators.validate_boolean("fasle")
        self.assertEqual("'fasle' is not a valid boolean value", msg)

    def test_validate_integer(self):
        msg = validators.validate_integer(1)
        self.assertIsNone(msg)
        msg = validators.validate_integer(0.1)
        self.assertEqual("'0.1' is not an integer", msg)
        msg = validators.validate_integer("1")
        self.assertIsNone(msg)
        msg = validators.validate_integer("0.1")
        self.assertEqual("'0.1' is not an integer", msg)
        msg = validators.validate_integer(True)
        self.assertEqual("'True' is not an integer:boolean", msg)
        msg = validators.validate_integer(False)
        self.assertEqual("'False' is not an integer:boolean", msg)
        msg = validators.validate_integer(float('Inf'))
        self.assertEqual("'inf' is not an integer", msg)
        msg = validators.validate_integer(None)
        self.assertEqual("'None' is not an integer", msg)

    def test_validate_integer_values(self):
        msg = validators.validate_integer(2, [2, 3, 4, 5])
        self.assertIsNone(msg)
        msg = validators.validate_integer(1, [2, 3, 4, 5])
        self.assertEqual("1 is not in valid_values", msg)

    def test_validate_no_whitespace(self):
        data = 'no_white_space'
        result = validators.validate_no_whitespace(data)
        self.assertEqual(data, result)

        self.assertRaises(n_exc.InvalidInput,
                          validators.validate_no_whitespace,
                          'i have whitespace')

        self.assertRaises(n_exc.InvalidInput,
                          validators.validate_no_whitespace,
                          'i\thave\twhitespace')

        for ws in string.whitespace:
            self.assertRaises(n_exc.InvalidInput,
                              validators.validate_no_whitespace,
                              '%swhitespace-at-head' % ws)
            self.assertRaises(n_exc.InvalidInput,
                              validators.validate_no_whitespace,
                              'whitespace-at-tail%s' % ws)

    def test_validate_range(self):
        msg = validators.validate_range(1, [1, 9])
        self.assertIsNone(msg)

        msg = validators.validate_range(5, [1, 9])
        self.assertIsNone(msg)

        msg = validators.validate_range(9, [1, 9])
        self.assertIsNone(msg)

        msg = validators.validate_range(1, (1, 9))
        self.assertIsNone(msg)

        msg = validators.validate_range(5, (1, 9))
        self.assertIsNone(msg)

        msg = validators.validate_range(9, (1, 9))
        self.assertIsNone(msg)

        msg = validators.validate_range(0, [1, 9])
        self.assertEqual("'0' is too small - must be at least '1'", msg)

        msg = validators.validate_range(10, (1, 9))
        self.assertEqual("'10' is too large - must be no larger than '9'", msg)

        msg = validators.validate_range("bogus", (1, 9))
        self.assertEqual("'bogus' is not an integer", msg)

        msg = validators.validate_range(10, (validators.UNLIMITED,
                                             validators.UNLIMITED))
        self.assertIsNone(msg)

        msg = validators.validate_range(10, (1, validators.UNLIMITED))
        self.assertIsNone(msg)

        msg = validators.validate_range(1, (validators.UNLIMITED, 9))
        self.assertIsNone(msg)

        msg = validators.validate_range(-1, (0, validators.UNLIMITED))
        self.assertEqual("'-1' is too small - must be at least '0'", msg)

        msg = validators.validate_range(10, (validators.UNLIMITED, 9))
        self.assertEqual("'10' is too large - must be no larger than '9'", msg)

    @mock.patch("neutron_lib.api.validators.validate_range")
    def test_validate_range_or_none(self, mock_validate_range):
        msg = validators.validate_range_or_none(None, [1, 9])
        self.assertFalse(mock_validate_range.called)
        self.assertIsNone(msg)

        validators.validate_range_or_none(1, [1, 9])
        mock_validate_range.assert_called_once_with(1, [1, 9])

    def _test_validate_mac_address(self, validator, allow_none=False):
        mac_addr = "ff:16:3e:4f:00:00"
        msg = validator(mac_addr)
        self.assertIsNone(msg)

        mac_addr = "ffa:16:3e:4f:00:00"
        msg = validator(mac_addr)
        err_msg = "'%s' is not a valid MAC address"
        self.assertEqual(err_msg % mac_addr, msg)

        for invalid_mac_addr in constants.INVALID_MAC_ADDRESSES:
            msg = validator(invalid_mac_addr)
            self.assertEqual(err_msg % invalid_mac_addr, msg)

        mac_addr = "123"
        msg = validator(mac_addr)
        self.assertEqual(err_msg % mac_addr, msg)

        mac_addr = None
        msg = validator(mac_addr)
        if allow_none:
            self.assertIsNone(msg)
        else:
            self.assertEqual(err_msg % mac_addr, msg)

        mac_addr = "ff:16:3e:4f:00:00\r"
        msg = validator(mac_addr)
        self.assertEqual(err_msg % mac_addr, msg)

    def test_validate_mac_address(self):
        self._test_validate_mac_address(validators.validate_mac_address)

    def test_validate_mac_address_or_none(self):
        self._test_validate_mac_address(
            validators.validate_mac_address_or_none, allow_none=True)

    def test_validate_ip_address(self):
        ip_addr = '1.1.1.1'
        msg = validators.validate_ip_address(ip_addr)
        self.assertIsNone(msg)

        ip_addr = '1111.1.1.1'
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        # Depending on platform to run UTs, this case might or might not be
        # an equivalent to test_validate_ip_address_bsd.
        ip_addr = '1' * 59
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        ip_addr = '1.1.1.1 has whitespace'
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        ip_addr = '111.1.1.1\twhitespace'
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        ip_addr = '111.1.1.1\nwhitespace'
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        for ws in string.whitespace:
            ip_addr = '%s111.1.1.1' % ws
            msg = validators.validate_ip_address(ip_addr)
            self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

        for ws in string.whitespace:
            ip_addr = '111.1.1.1%s' % ws
            msg = validators.validate_ip_address(ip_addr)
            self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

    def test_validate_ip_address_with_leading_zero(self):
        ip_addr = '1.1.1.01'
        expected_msg = ("'%(data)s' is not an accepted IP address, "
                        "'%(ip)s' is recommended")
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual(expected_msg % {"data": ip_addr, "ip": '1.1.1.1'},
                         msg)

        ip_addr = '1.1.1.011'
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual(expected_msg % {"data": ip_addr, "ip": '1.1.1.11'},
                         msg)

        ip_addr = '1.1.1.09'
        msg = validators.validate_ip_address(ip_addr)
        self.assertEqual(expected_msg % {"data": ip_addr, "ip": '1.1.1.9'},
                         msg)

        ip_addr = "fe80:0:0:0:0:0:0:0001"
        msg = validators.validate_ip_address(ip_addr)
        self.assertIsNone(msg)

    def test_validate_ip_address_bsd(self):
        # NOTE(yamamoto):  On NetBSD and OS X, netaddr.IPAddress() accepts
        # '1' * 59 as a valid address.  The behaviour is inherited from
        # libc behaviour there.  This test ensures that our validator reject
        # such addresses on such platforms by mocking netaddr to emulate
        # the behaviour.
        ip_addr = '1' * 59
        with mock.patch('netaddr.IPAddress') as ip_address_cls:
            msg = validators.validate_ip_address(ip_addr)
        ip_address_cls.assert_called_once_with(ip_addr,
                                               flags=netaddr.core.ZEROFILL)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

    def test_validate_ip_pools(self):
        pools = [[{'end': '10.0.0.254'}],
                 [{'start': '10.0.0.254'}],
                 [{'start': '1000.0.0.254',
                   'end': '1.1.1.1'}],
                 [{'start': '10.0.0.2', 'end': '10.0.0.254',
                   'forza': 'juve'}],
                 [{'start': '10.0.0.2', 'end': '10.0.0.254'},
                  {'end': '10.0.0.254'}],
                 [None],
                 None]
        for pool in pools:
            msg = validators.validate_ip_pools(pool)
            self.assertIsNotNone(msg)

        pools = [[{'end': '10.0.0.254', 'start': '10.0.0.2'},
                  {'start': '11.0.0.2', 'end': '11.1.1.1'}],
                 [{'start': '11.0.0.2', 'end': '11.0.0.100'}]]
        for pool in pools:
            msg = validators.validate_ip_pools(pool)
            self.assertIsNone(msg)

        invalid_ip = '10.0.0.2\r'
        pools = [[{'end': '10.0.0.254', 'start': invalid_ip}]]
        for pool in pools:
            msg = validators.validate_ip_pools(pool)
            self.assertEqual("'%s' is not a valid IP address" % invalid_ip,
                             msg)

    def test_validate_fixed_ips(self):
        fixed_ips = [
            {'data': [{'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1111.1.1.1'}],
             'error_msg': "'1111.1.1.1' is not a valid IP address"},
            {'data': [{'subnet_id': 'invalid',
                       'ip_address': '1.1.1.1'}],
             'error_msg': "'invalid' is not a valid UUID"},
            {'data': None,
             'error_msg': "Invalid data format for fixed IP: 'None'"},
            {'data': "1.1.1.1",
             'error_msg': "Invalid data format for fixed IP: '1.1.1.1'"},
            {'data': ['00000000-ffff-ffff-ffff-000000000000', '1.1.1.1'],
             'error_msg': "Invalid data format for fixed IP: "
                          "'00000000-ffff-ffff-ffff-000000000000'"},
            {'data': [['00000000-ffff-ffff-ffff-000000000000', '1.1.1.1']],
             'error_msg': "Invalid data format for fixed IP: "
                          "'['00000000-ffff-ffff-ffff-000000000000', "
                          "'1.1.1.1']'"},
            {'data': [{'subnet_id': '00000000-0fff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'},
                      {'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'}],
             'error_msg': "Duplicate IP address '1.1.1.1'"}]
        for fixed in fixed_ips:
            msg = validators.validate_fixed_ips(fixed['data'])
            self.assertEqual(fixed['error_msg'], msg)

        fixed_ips = [[{'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'}],
                     [{'subnet_id': '00000000-0fff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.1'},
                      {'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                       'ip_address': '1.1.1.2'}]]
        for fixed in fixed_ips:
            msg = validators.validate_fixed_ips(fixed)
            self.assertIsNone(msg)

    def test_validate_nameservers(self):
        ns_pools = [['1.1.1.2', '1.1.1.2'],
                    ['www.hostname.com', 'www.hostname.com'],
                    ['1000.0.0.1'],
                    ['www.hostname.com'],
                    ['www.great.marathons.to.travel'],
                    ['valid'],
                    ['77.hostname.com'],
                    ['1' * 59],
                    ['www.internal.hostname.com'],
                    None]

        for ns in ns_pools:
            msg = validators.validate_nameservers(ns, None)
            self.assertIsNotNone(msg)

        ns_pools = [['100.0.0.2'],
                    ['1.1.1.1', '1.1.1.2']]

        for ns in ns_pools:
            msg = validators.validate_nameservers(ns, None)
            self.assertIsNone(msg)

    def test_validate_hostroutes(self):
        hostroute_pools = [[{'destination': '100.0.0.0/24'}],
                           [{'nexthop': '10.0.2.20'}],
                           [{'nexthop': '10.0.2.20',
                             'forza': 'juve',
                             'destination': '100.0.0.0/8'}],
                           [{'nexthop': '1110.0.2.20',
                             'destination': '100.0.0.0/8'}],
                           [{'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0'}],
                           [{'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0/8'},
                            {'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0/8'}],
                           [None],
                           None]
        for host_routes in hostroute_pools:
            msg = validators.validate_hostroutes(host_routes, None)
            self.assertIsNotNone(msg)

        hostroute_pools = [[{'destination': '100.0.0.0/24',
                             'nexthop': '10.0.2.20'}],
                           [{'nexthop': '10.0.2.20',
                             'destination': '100.0.0.0/8'},
                            {'nexthop': '10.0.2.20',
                             'destination': '101.0.0.0/8'}]]
        for host_routes in hostroute_pools:
            msg = validators.validate_hostroutes(host_routes, None)
            self.assertIsNone(msg)

    def test_validate_ip_address_or_none(self):
        ip_addr = None
        msg = validators.validate_ip_address_or_none(ip_addr)
        self.assertIsNone(msg)

        ip_addr = '1.1.1.1'
        msg = validators.validate_ip_address_or_none(ip_addr)
        self.assertIsNone(msg)

        ip_addr = '1111.1.1.1'
        msg = validators.validate_ip_address_or_none(ip_addr)
        self.assertEqual("'%s' is not a valid IP address" % ip_addr, msg)

    def test_uuid_pattern(self):
        data = 'garbage'
        msg = validators.validate_regex(data, constants.UUID_PATTERN)
        self.assertIsNotNone(msg)

        data = '00000000-ffff-ffff-ffff-000000000000'
        msg = validators.validate_regex(data, constants.UUID_PATTERN)
        self.assertIsNone(msg)

    def test_mac_pattern(self):
        # Valid - 3 octets
        base_mac = "fa:16:3e:00:00:00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNone(msg)

        # Valid - 4 octets
        base_mac = "fa:16:3e:4f:00:00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNone(msg)

        # Invalid - not unicast
        base_mac = "01:16:3e:4f:00:00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "a:16:3e:4f:00:00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "ffa:16:3e:4f:00:00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "01163e4f0000"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "01-16-3e-4f-00-00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "00:16:3:f:00:00"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

        # Invalid - invalid format
        base_mac = "12:3:4:5:67:89ab"
        msg = validators.validate_regex(base_mac, validators.MAC_PATTERN)
        self.assertIsNotNone(msg)

    def _test_validate_subnet(self, validator, allow_none=False):
        # Valid - IPv4
        cidr = "10.0.2.0/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - IPv6 without final octets
        cidr = "fe80::/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - IPv6 with final octets
        cidr = "fe80::/24"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - uncompressed ipv6 address
        cidr = "fe80:0:0:0:0:0:0:0/128"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - ipv6 address with multiple consecutive zero
        cidr = "2001:0db8:0:0:1::1/128"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - ipv6 address with multiple consecutive zero
        cidr = "2001:0db8::1:0:0:1/128"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Valid - ipv6 address with multiple consecutive zero
        cidr = "2001::0:1:0:0:1100/120"
        msg = validator(cidr, None)
        self.assertIsNone(msg)

        # Invalid - abbreviated ipv4 address
        cidr = "10/24"
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

        # Invalid - IPv4 missing mask
        cidr = "10.0.2.0"
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

        # Valid - IPv4 with non-zero masked bits is ok
        for i in range(1, 255):
            cidr = "192.168.1.%s/24" % i
            msg = validator(cidr, None)
            self.assertIsNone(msg)

        # Invalid - IPv6 without final octets, missing mask
        cidr = "fe80::"
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

        # Invalid - IPv6 with final octets, missing mask
        cidr = "fe80::0"
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

        # Invalid - Address format error
        cidr = 'invalid'
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

        cidr = None
        msg = validator(cidr, None)
        if allow_none:
            self.assertIsNone(msg)
        else:
            error = "'%s' is not a valid IP subnet" % cidr
            self.assertEqual(error, msg)

        # Invalid - IPv4 with trailing CR
        cidr = "10.0.2.0/24\r"
        msg = validator(cidr, None)
        error = "'%s' is not a valid IP subnet" % cidr
        self.assertEqual(error, msg)

    def test_validate_subnet(self):
        self._test_validate_subnet(validators.validate_subnet)

    def test_validate_route_cidr(self):
        # Valid - CIDR
        cidr = "10.0.0.0/8"
        msg = validators.validate_route_cidr(cidr, None)
        self.assertIsNone(msg)

        # Valid - CIDR
        cidr = "192.168.1.1/32"
        msg = validators.validate_route_cidr(cidr, None)
        self.assertIsNone(msg)

        # Invalid - CIDR
        cidr = "192.168.1.1/8"
        msg = validators.validate_route_cidr(cidr, None)
        error = "'%s' is not a valid CIDR" % cidr
        self.assertEqual(error, msg)

        # Invalid - loopback CIDR
        cidr = "127.0.0.0/8"
        msg = validators.validate_route_cidr(cidr, None)
        error = _("'%(data)s' is not a routable CIDR") % {"data": cidr}
        self.assertEqual(error, msg)

        # Invalid - CIDR format error
        cidr = 'invalid'
        msg = validators.validate_route_cidr(cidr, None)
        error = "'%s' is not a valid CIDR" % cidr
        self.assertEqual(error, msg)

    def test_validate_subnet_or_none(self):
        self._test_validate_subnet(validators.validate_subnet_or_none,
                                   allow_none=True)

    def test_validate_subnet_list(self):
        msg = validators.validate_subnet_list('abc')
        self.assertEqual("'abc' is not a list", msg)
        msg = validators.validate_subnet_list(['10.1.0.0/24',
                                               '10.2.0.0/24',
                                               '10.1.0.0/24'])
        self.assertEqual("Duplicate items in the list: '10.1.0.0/24'", msg)

        cidrs = ['10.1.0.0/24', '10.2.0.0']
        msg = validators.validate_subnet_list(cidrs)
        error = "'%s' is not a valid IP subnet" % cidrs[1]
        self.assertEqual(error, msg)

    def _test_validate_regex(self, validator, allow_none=False):
        pattern = '[hc]at'

        data = None
        msg = validator(data, pattern)
        if allow_none:
            self.assertIsNone(msg)
        else:
            self.assertEqual("'None' is not a valid input", msg)

        data = 'bat'
        msg = validator(data, pattern)
        self.assertEqual("'%s' is not a valid input" % data, msg)

        data = 'hat'
        msg = validator(data, pattern)
        self.assertIsNone(msg)

        data = 'cat'
        msg = validator(data, pattern)
        self.assertIsNone(msg)

    def test_validate_regex(self):
        self._test_validate_regex(validators.validate_regex)

    def test_validate_regex_or_none(self):
        self._test_validate_regex(validators.validate_regex_or_none,
                                  allow_none=True)

    def test_validate_list_of_regex_or_none(self):
        pattern = '[hc]at|^$'

        list_of_regex = ['hat', 'cat', '']
        msg = validators.validate_list_of_regex_or_none(list_of_regex, pattern)
        self.assertIsNone(msg)

        list_of_regex = ['bat', 'hat', 'cat', '']
        msg = validators.validate_list_of_regex_or_none(list_of_regex, pattern)
        self.assertEqual("'bat' is not a valid input", msg)

        empty_list = []
        msg = validators.validate_list_of_regex_or_none(empty_list, pattern)

    def test_validate_subnetpool_id(self):
        msg = validators.validate_subnetpool_id(constants.IPV6_PD_POOL_ID)
        self.assertIsNone(msg)

        msg = validators.validate_subnetpool_id(
            '00000000-ffff-ffff-ffff-000000000000')
        self.assertIsNone(msg)

    def test_validate_subnetpool_id_or_none(self):
        msg = validators.validate_subnetpool_id_or_none(None)
        self.assertIsNone(msg)

        msg = validators.validate_subnetpool_id_or_none(
            '00000000-ffff-ffff-ffff-000000000000')
        self.assertIsNone(msg)

    def test_validate_uuid(self):
        invalid_uuids = [None,
                         123,
                         '123',
                         't5069610-744b-42a7-8bd8-ceac1a229cd4',
                         'e5069610-744bb-42a7-8bd8-ceac1a229cd4']
        for uuid in invalid_uuids:
            msg = validators.validate_uuid(uuid)
            error = "'%s' is not a valid UUID" % uuid
            self.assertEqual(error, msg)

        msg = validators.validate_uuid('00000000-ffff-ffff-ffff-000000000000')
        self.assertIsNone(msg)

    def test_validate_uuid_list(self):
        bad_uuid_list = ['00000000-ffff-ffff-ffff-000000000000',
                         '00000000-ffff-ffff-ffff-000000000001',
                         '123']
        msg = validators.validate_uuid_list(bad_uuid_list,
                                            valid_values='parameter not used')
        error = "'%s' is not a valid UUID" % bad_uuid_list[2]
        self.assertEqual(error, msg)

        good_uuid_list = ['00000000-ffff-ffff-ffff-000000000000',
                          '00000000-ffff-ffff-ffff-000000000001']
        msg = validators.validate_uuid_list(good_uuid_list,
                                            valid_values='parameter not used')
        self.assertIsNone(msg)

    def test__validate_list_of_items(self):
        # check not a list
        items = [None,
                 123,
                 'e5069610-744b-42a7-8bd8-ceac1a229cd4',
                 '12345678123456781234567812345678',
                 {'uuid': 'e5069610-744b-42a7-8bd8-ceac1a229cd4'}]
        for item in items:
            msg = validators._validate_list_of_items(mock.Mock(), item)
            error = "'%s' is not a list" % item
            self.assertEqual(error, msg)

        # check duplicate items in a list
        duplicate_items = ['e5069610-744b-42a7-8bd8-ceac1a229cd4',
                           'f3eeab00-8367-4524-b662-55e64d4cacb5',
                           'e5069610-744b-42a7-8bd8-ceac1a229cd4']
        msg = validators._validate_list_of_items(mock.Mock(), duplicate_items)
        error = ("Duplicate items in the list: "
                 "'e5069610-744b-42a7-8bd8-ceac1a229cd4'")
        self.assertEqual(error, msg)

        # check valid lists
        valid_lists = [[],
                       [1, 2, 3],
                       ['a', 'b', 'c']]
        for list_obj in valid_lists:
            msg = validators._validate_list_of_items(
                mock.Mock(return_value=None), list_obj)
            self.assertIsNone(msg)

    def test__test__validate_list_of_items_non_empty(self):
        items = None
        msg = validators._validate_list_of_items_non_empty(mock.Mock(), items)
        error = "'%s' is not a list" % items
        self.assertEqual(error, msg)

        items = []
        msg = validators._validate_list_of_items_non_empty(mock.Mock(), items)
        self.assertEqual("List should not be empty", msg)

    def test__validate_list_of_items_collect_duplicates(self):
        items = ['a', 'b', 'duplicate_1', 'duplicate_2', 'duplicate_1',
                 'duplicate_2', 'duplicate_2', 'c']
        msg = validators._validate_list_of_items(mock.Mock(), items)
        error = ("Duplicate items in the list: '%s'"
                 % 'duplicate_1, duplicate_2')
        self.assertEqual(error, msg)

        items = [['a', 'b'], ['c', 'd'], ['a', 'b']]
        msg = validators._validate_list_of_items(mock.Mock(), items)
        error = "Duplicate items in the list: '%s'" % str(['a', 'b'])
        self.assertEqual(error, msg)

        items = [{'a': 'b'}, {'c': 'd'}, {'a': 'b'}]
        msg = validators._validate_list_of_items(mock.Mock(), items)
        error = "Duplicate items in the list: '%s'" % str({'a': 'b'})
        self.assertEqual(error, msg)

    def test_validate_dict_type(self):
        for value in (None, True, '1', []):
            self.assertEqual("'%s' is not a dictionary" % value,
                             validators.validate_dict(value))

    def test_validate_dict_without_constraints(self):
        msg = validators.validate_dict({})
        self.assertIsNone(msg)

        # Validate a dictionary without constraints.
        msg = validators.validate_dict({'key': 'value'})
        self.assertIsNone(msg)

    def test_validate_a_valid_dict_with_constraints(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        msg = validators.validate_dict(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_dict_with_invalid_validator(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        constraints['key1'] = {'type:unsupported': None, 'required': True}
        msg = validators.validate_dict(dictionary, constraints)
        self.assertEqual("Validator 'type:unsupported' does not exist", msg)

    def test_validate_dict_not_required_keys(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        del dictionary['key2']
        msg = validators.validate_dict(dictionary, constraints)
        self.assertIsNone(msg, 'Field that was not required by the specs was'
                               'required by the validator.')

    def test_validate_dict_required_keys(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        del dictionary['key1']
        msg = validators.validate_dict(dictionary, constraints)
        self.assertIn('Expected keys:', msg)

    def test_validate_dict_wrong_values(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        dictionary['key1'] = 'UNSUPPORTED'
        msg = validators.validate_dict(dictionary, constraints)
        self.assertIsNotNone(msg)

    def test_validate_dict_unexpected_keys(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        dictionary['unexpected_key'] = 'val'
        msg = validators.validate_dict(dictionary, constraints)
        self.assertIn('Unexpected keys supplied:', msg)

    def test_validate_dict_convert_boolean(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        constraints['key_bool'] = {
            'type:boolean': None,
            'required': False,
            'convert_to': converters.convert_to_boolean}
        dictionary['key_bool'] = 'true'
        msg = validators.validate_dict(dictionary, constraints)
        self.assertIsNone(msg)
        # Explicitly comparing with literal 'True' as assertTrue
        # succeeds also for 'true'
        self.assertIs(True, dictionary['key_bool'])

    def test_subdictionary(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        del dictionary['key3']['k4']
        dictionary['key3']['k5'] = 'a string value'
        msg = validators.validate_dict(dictionary, constraints)
        self.assertIn('Expected keys:', msg)

    def test_validate_dict_or_none(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        # Check whether None is a valid value.
        msg = validators.validate_dict_or_none(None, constraints)
        self.assertIsNone(msg, 'Validation of a None dictionary failed.')

        # Check validation of a regular dictionary.
        msg = validators.validate_dict_or_none(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_dict_or_empty(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        # Check whether an empty dictionary is valid.
        msg = validators.validate_dict_or_empty({}, constraints)
        self.assertIsNone(msg, 'Validation of a None dictionary failed.')

        # Check validation of a regular dictionary.
        msg = validators.validate_dict_or_empty(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_dict_or_nodata(self):
        dictionary, constraints = self._construct_dict_and_constraints()

        # Check whether no data is a valid value.
        msg = validators.validate_dict_or_nodata(None, constraints)
        self.assertIsNone(msg, 'Validation of None for no-data failed.')
        msg = validators.validate_dict_or_nodata({}, constraints)
        self.assertIsNone(msg, 'Validation of empty dict for no-data failed.')

        # Check validation of a regular dictionary.
        msg = validators.validate_dict_or_nodata(dictionary, constraints)
        self.assertIsNone(msg, 'Validation of a valid dictionary failed.')

    def test_validate_non_negative(self):
        msg = validators.validate_non_negative('abc')
        self.assertEqual("'abc' is not an integer", msg)

        for value in (-1, '-2'):
            self.assertEqual("'%s' should be non-negative" % value,
                             validators.validate_non_negative(value))

        for value in (0, 1, '2', True, False):
            msg = validators.validate_non_negative(value)
            self.assertIsNone(msg)

    def test_validate_subports_invalid_body(self):
        self.assertIsNotNone(validators.validate_subports(None))

    def test_validate_subports_invalid_subport_object(self):
        self.assertIsNotNone(validators.validate_subports(['foo_port']))

    def test_validate_subports_invalid_port_uuid(self):
        body = [{'port_id': 'foo_port'}]
        self.assertIsNotNone(validators.validate_subports(body))

    def test_validate_subports_invalid_missing_port_id(self):
        body = [{'poort_id': 'foo_port'}]
        self.assertIsNotNone(validators.validate_subports(body))

    def test_validate_subports_invalid_duplicate_port_ids(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000'},
            {'port_id': '00000000-ffff-ffff-ffff-000000000000'}
        ]
        self.assertIsNotNone(validators.validate_subports(body))

    def test_validate_subports_invalid_incomplete_segmentation_details(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_id': '3'}
        ]
        self.assertIsNotNone(validators.validate_subports(body))

    def test_validate_subports_invalid_unknown_paramenter(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_id': '3', 'segmeNAtion_type': 'vlan'}
        ]
        self.assertIsNotNone(validators.validate_subports(body))

    def test_validate_subports_invalid_duplicate_segmentation_id(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_id': 0, 'segmentation_type': 'vlan'},
            {'port_id': '11111111-ffff-ffff-ffff-000000000000',
             'segmentation_id': 0, 'segmentation_type': 'vlan'}
        ]
        self.assertIsNotNone(validators.validate_subports(body))

    def test_validate_subports_with_segmentation_id_0(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_id': '0', 'segmentation_type': 'vlan'}
        ]
        self.assertIsNone(validators.validate_subports(body))

    def test_validate_subports_inherit_segmentation_details(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_type': 'inherit'}
        ]
        self.assertIsNone(validators.validate_subports(body))

    def test_validate_subports_valid_unique_segmentation_id(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_id': '3', 'segmentation_type': 'vlan'},
            {'port_id': '11111111-ffff-ffff-ffff-000000000000',
             'segmentation_id': '3', 'segmentation_type': 'vxlan'}
        ]
        self.assertIsNone(validators.validate_subports(body))

    def test_validate_subports_valid_empty_body(self):
        self.assertIsNone(validators.validate_subports([]))

    def test_validate_subports_valid_subports_with_segmentation_details(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000',
             'segmentation_id': '3', 'segmentation_type': 'vlan'},
            {'port_id': '11111111-ffff-ffff-ffff-000000000000',
             'segmentation_id': '5', 'segmentation_type': 'vlan'}
        ]
        self.assertIsNone(validators.validate_subports(body))

    def test_validate_subports_valid_subports(self):
        body = [
            {'port_id': '00000000-ffff-ffff-ffff-000000000000'},
            {'port_id': '11111111-ffff-ffff-ffff-000000000000'},
        ]
        self.assertIsNone(validators.validate_subports(body))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_valid_string_new(self, CONF):
        CONF.sg_filter_ethertypes = True
        self.assertIsNone(validators.validate_ethertype('IPv4'))
        self.assertIsNone(validators.validate_ethertype('IPv6'))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_valid_string_old(self, CONF):
        CONF.sg_filter_ethertypes = False
        self.assertIsNone(validators.validate_ethertype('IPv4'))
        self.assertIsNone(validators.validate_ethertype('IPv6'))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_invalid_string(self, CONF):
        CONF.sg_filter_ethertypes = False
        self.assertEqual(('Ethertype 0x4008 is not a valid ethertype, must be '
                          'one of IPv4,IPv6.'),
                         validators.validate_ethertype('0x4008'))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_extended_valid_string(self, CONF):
        CONF.sg_filter_ethertypes = True
        self.assertIsNone(validators.validate_ethertype('0x4008'))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_extended_valid_hexint(self, CONF):
        CONF.sg_filter_ethertypes = True
        self.assertIsNone(validators.validate_ethertype(0x4008))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_extended_invalid_negative(self, CONF):
        CONF.sg_filter_ethertypes = True
        self.assertEqual(("Ethertype -16392 is not a two octet "
                          "hexadecimal number or ethertype name."),
                         validators.validate_ethertype('-0x4008'))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_extended_invalid_nonhex(self, CONF):
        CONF.sg_filter_ethertypes = True
        self.assertEqual(("Ethertype invalid is not a two octet "
                          "hexadecimal number or ethertype name."),
                         validators.validate_ethertype('invalid'))

    @mock.patch('oslo_config.cfg.CONF')
    def test_validate_ethertype_extended_invalid_toobig(self, CONF):
        CONF.sg_filter_ethertypes = True
        self.assertEqual(("Ethertype 3735928559 is not a two octet "
                          "hexadecimal number or ethertype name."),
                         validators.validate_ethertype('0xdeadbeef'))


class TestValidateIPSubnetNone(base.BaseTestCase):

    def test_validate_none(self):
        self.assertIsNone(validators.validate_ip_or_subnet_or_none(None))

    def test_validate_ipv4(self):
        testdata = "172.0.0.1"
        self.assertIsNone(validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv4_subnet(self):
        testdata = "172.0.0.1/24"
        self.assertIsNone(validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv6(self):
        testdata = "2001:0db8:0a0b:12f0:0000:0000:0000:0001"
        self.assertIsNone(validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv6_subnet(self):
        testdata = "::1/128"
        self.assertIsNone(validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv4_invalid(self):
        testdata = "300.0.0.1"
        self.assertEqual(("'300.0.0.1' is neither a valid IP address, nor is "
                          "it a valid IP subnet"),
                         validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv4_subnet_invalid(self):
        testdata = "172.0.0.1/45"
        self.assertEqual(("'172.0.0.1/45' is neither a valid IP address, nor "
                          "is it a valid IP subnet"),
                         validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv6_invalid(self):
        testdata = "xxxx:0db8:0a0b:12f0:0000:0000:0000:0001"
        self.assertEqual(("'xxxx:0db8:0a0b:12f0:0000:0000:0000:0001' is "
                          "neither a valid IP address, nor is it a valid IP "
                          "subnet"),
                         validators.validate_ip_or_subnet_or_none(testdata))

    def test_validate_ipv6_subnet_invalid(self):
        testdata = "::1/2048"
        self.assertEqual(("'::1/2048' is neither a valid IP address, nor is "
                          "it a valid IP subnet"),
                         validators.validate_ip_or_subnet_or_none(testdata))


class TestPortRangeValidation(base.BaseTestCase):

    def test_valid_port_specific_range(self):
        result = validators.validate_port_range_or_none("4:5",
                                                        [1, 65535])
        self.assertIsNone(result)

    def test_invalid_port_specific_range(self):
        result = validators.validate_port_range_or_none("4:500000",
                                                        [1, 65535])
        self.assertEqual("Invalid port: 500000", result)

    def test_invalid_port_for_specific_range(self):
        result = validators.validate_port_range_or_none("0:10",
                                                        [1, 65535])
        self.assertEqual("Invalid port: 0, the port must be in"
                         " the range [1, 65535]", result)

    def test_valid_port(self):
        result = validators.validate_port_range_or_none("80")
        self.assertIsNone(result)

    def test_valid_port_integer(self):
        result = validators.validate_port_range_or_none(80)
        self.assertIsNone(result)

    def test_valid_range(self):
        # NOTE(huntxu): This case would fail when ports are compared as
        # strings, since '9' > '1111'.
        result = validators.validate_port_range_or_none("9:1111")
        self.assertIsNone(result)

    def test_port_too_high(self):
        result = validators.validate_port_range_or_none("99999")
        self.assertEqual("Invalid port: 99999", result)

    def test_port_too_low(self):
        result = validators.validate_port_range_or_none("-1")
        self.assertEqual("Invalid port: -1", result)

    def test_range_too_high(self):
        result = validators.validate_port_range_or_none("80:99999")
        self.assertEqual("Invalid port: 99999", result)

    def test_range_too_low(self):
        result = validators.validate_port_range_or_none("-1:8888")
        self.assertEqual("Invalid port: -1", result)

    def test_range_wrong_way(self):
        # NOTE(huntxu): This case would fail when ports are compared as
        # strings, since '1111' < '9'.
        result = validators.validate_port_range_or_none("1111:9")
        self.assertEqual("First port in a port range must be lower than the "
                         "second port", result)

    def test_range_invalid(self):
        result = validators.validate_port_range_or_none("DEAD:BEEF")
        self.assertEqual("Invalid port: DEAD", result)

    def test_range_bad_input(self):
        result = validators.validate_port_range_or_none(['a', 'b', 'c'])
        self.assertEqual("Invalid port: ['a', 'b', 'c']", result)

    def test_range_colon(self):
        result = validators.validate_port_range_or_none(":")
        self.assertEqual("Port range must be two integers separated by a "
                         "colon", result)

    def test_too_many_colons(self):
        result = validators.validate_port_range_or_none("80:888:8888")
        self.assertEqual("Port range must be two integers separated by a "
                         "colon", result)


class TestAnyKeySpecs(base.BaseTestCase):

    def test_data_is_none(self):
        self.assertIsNone(
            validators.validate_any_key_specs_or_none(None, key_specs={}))

    def test_data_is_not_list(self):
        for t in [dict(), set(), 'abc', 1, True]:
            self.assertRaises(
                n_exc.InvalidInput,
                validators.validate_any_key_specs_or_none, t, key_specs={})

    def test_data_invalid_keys(self):
        data = [{'opt_name': 'a', 'opt_value': 'A'},
                {'opt_name': 'b', 'opt_valuee': 'B'}]
        self.assertRaisesRegex(
            n_exc.InvalidInput,
            "No valid key specs",
            validators.validate_any_key_specs_or_none,
            data, key_specs=extra_dhcp_opt.EXTRA_DHCP_OPT_KEY_SPECS)

    def test_data_optional_key(self):
        data = [{'opt_name': 'a', 'opt_value': 'A'},
                {'opt_name': 'b', 'opt_value': 'B', 'ip_version': '4'}]
        self.assertIsNone(
            validators.validate_any_key_specs_or_none(
                data, key_specs=extra_dhcp_opt.EXTRA_DHCP_OPT_KEY_SPECS))

    def test_data_optional_key_invalid(self):
        data = [{'opt_name': 'a', 'opt_value': 'A'},
                {'opt_name': 'b', 'opt_value': 'B', 'ip_version': '3'}]
        self.assertRaisesRegex(
            n_exc.InvalidInput,
            "No valid key specs",
            validators.validate_any_key_specs_or_none,
            data, key_specs=extra_dhcp_opt.EXTRA_DHCP_OPT_KEY_SPECS)

    def test_data_conditional_spec(self):
        data = [{'opt_name': 'router', 'opt_value': None},
                {'opt_name': 'b', 'opt_value': 'B', 'ip_version': '4'}]
        self.assertIsNone(
            validators.validate_any_key_specs_or_none(
                data, key_specs=extra_dhcp_opt.EXTRA_DHCP_OPT_KEY_SPECS))


class TestServicePluginType(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self._plugins = directory._PluginDirectory()
        self._plugins.add_plugin('stype', mock.Mock())
        self.useFixture(fixture.PluginDirectoryFixture(
            plugin_directory=self._plugins))

    def test_valid_plugin_type(self):
        self.assertIsNone(validators.validate_service_plugin_type('stype'))

    def test_invalid_plugin_type(self):
        self.assertRaisesRegex(
            n_exc.InvalidServiceType,
            'Invalid service type',
            validators.validate_service_plugin_type, 'ntype')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.plugins import directory
from neutron_lib.tests import _base as base


def fake_plugin():
    pass


class DirectoryTestCase(base.BaseTestCase):

    def test__create_plugin_directory(self):
        self.assertIsNotNone(directory._create_plugin_directory())

    def test__get_plugin_directory(self):
        self.assertIsNotNone(directory._get_plugin_directory())

    def test_add_plugin(self):
        directory.add_plugin('foo', fake_plugin)
        self.assertIn('foo', directory.get_plugins())

    def test_get_plugin_core_none(self):
        self.assertIsNone(directory.get_plugin())

    def test_get_plugin_alias_none(self):
        self.assertIsNone(directory.get_plugin('foo'))

    def test_get_plugin_core(self):
        directory.add_plugin('CORE', fake_plugin)
        self.assertIsNotNone(directory.get_plugin())

    def test_get_plugin_alias(self):
        directory.add_plugin('foo', fake_plugin)
        self.assertIsNotNone(directory.get_plugin('foo'))

    def test_get_plugins_none(self):
        self.assertFalse(directory.get_plugins())

    def test_get_unique_plugins_none(self):
        self.assertFalse(directory.get_unique_plugins())

    def test_get_plugins(self):
        directory.add_plugin('CORE', fake_plugin)
        self.assertIsNotNone(directory.get_plugins())

    def test_get_unique_plugins(self):
        directory.add_plugin('foo1', fake_plugin)
        directory.add_plugin('foo2', fake_plugin)
        self.assertEqual(1, len(directory.get_unique_plugins()))

    def test_is_loaded(self):
        self.assertFalse(directory.is_loaded())
        directory.add_plugin('foo1', fake_plugin)
        self.assertTrue(directory.is_loaded())


class PluginDirectoryTestCase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.plugin_directory = directory._PluginDirectory()

    def test_add_plugin(self):
        self.plugin_directory.add_plugin('foo', 'bar')
        self.assertEqual(1, len(self.plugin_directory._plugins))

    def test_get_plugin_not_found(self):
        self.assertIsNone(self.plugin_directory.get_plugin('foo'))

    def test_get_plugin_found(self):
        self.plugin_directory._plugins = {'foo': lambda *x, **y: 'bar'}
        plugin = self.plugin_directory.get_plugin('foo')
        self.assertEqual('bar', plugin())

    def test_plugins(self):
        self.plugin_directory._plugins = {'foo': lambda *x, **y: 'bar'}
        self.assertIsNotNone(self.plugin_directory.plugins)

    def test_unique_plugins(self):
        self.plugin_directory._plugins = {
            'foo1': fake_plugin,
            'foo2': fake_plugin,
        }
        self.assertEqual(1, len(self.plugin_directory.unique_plugins))

    def test_is_loaded(self):
        self.assertFalse(self.plugin_directory.is_loaded)
        self.plugin_directory._plugins = {'foo': lambda *x, **y: 'bar'}
        self.assertTrue(self.plugin_directory.is_loaded)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy

from neutron_lib import exceptions


def convert_filters(**kwargs):
    result = copy.deepcopy(kwargs)
    if 'tenant_id' in result:
        if 'project_id' in result:
            raise exceptions.TenantIdProjectIdFilterConflict()

        result['project_id'] = result.pop('tenant_id')
    return result


class FilterObj(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def filter(self, column):
        pass


class StringMatchingFilterObj(FilterObj):
    @property
    def is_contains(self):
        return bool(getattr(self, "contains", False))

    @property
    def is_starts(self):
        return bool(getattr(self, "starts", False))

    @property
    def is_ends(self):
        return bool(getattr(self, "ends", False))


class StringContains(StringMatchingFilterObj):

    def __init__(self, matching_string):
        super().__init__()
        self.contains = matching_string

    def filter(self, column):
        return column.contains(self.contains)


class StringStarts(StringMatchingFilterObj):

    def __init__(self, matching_string):
        super().__init__()
        self.starts = matching_string

    def filter(self, column):
        return column.startswith(self.starts)


class StringEnds(StringMatchingFilterObj):

    def __init__(self, matching_string):
        super().__init__()
        self.ends = matching_string

    def filter(self, column):
        return column.endswith(self.ends)


class NotIn(FilterObj):

    def __init__(self, value):
        super().__init__()
        self.value = value

    def filter(self, column):
        return ~column.in_(self.value)


class NotEqual(FilterObj):

    def __init__(self, value):
        super().__init__()
        self.value = value

    def filter(self, column):
        return column != self.value


def get_updatable_fields(cls, fields):
    fields = fields.copy()
    for field in cls.fields_no_update:
        if field in fields:
            del fields[field]
    return fields



# Copyright (c) 2020 Troila.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api.definitions import l3
from neutron_lib.api.definitions import l3_ext_gw_mode
from neutron_lib import constants


ALIAS = 'l3-ext-ndp-proxy'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Router support to enable NDP proxy'
API_PREFIX = ''
DESCRIPTION = 'Router extension to enable ndp proxy'
UPDATED_TIMESTAMP = '2020-08-25T00:00:00-00:00'
RESOURCE_NAME = l3.ROUTER
COLLECTION_NAME = l3.ROUTERS
ENABLE_NDP_PROXY = 'enable_ndp_proxy'
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        ENABLE_NDP_PROXY: {
            'allow_post': True,
            'allow_put': True,
            'convert_to': converters.convert_to_boolean_if_not_none,
            'default': constants.ATTR_NOT_SPECIFIED,
            'is_visible': True,
            'is_filter': True
        }
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [l3.ALIAS, l3_ext_gw_mode.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import netaddr
from oslo_db import exception
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa

from neutron_lib import context
from neutron_lib.db import sqlalchemytypes
from neutron_lib.tests import _base as test_base
from neutron_lib.tests import tools
from neutron_lib.utils import net


class SqlAlchemyTypesBaseTestCase(test_fixtures.OpportunisticDBTestMixin,
                                  test_base.BaseTestCase,
                                  metaclass=abc.ABCMeta):
    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        meta = sa.MetaData()
        self.test_table = self._get_test_table(meta)
        meta.create_all(self.engine)
        self.addCleanup(meta.drop_all, self.engine)
        self.ctxt = context.get_admin_context()

    @abc.abstractmethod
    def _get_test_table(self, meta):
        """Returns a new sa.Table() object for this test case."""

    def _add_row(self, **kargs):
        row_insert = self.test_table.insert().values(**kargs)
        with self.engine.connect() as conn, conn.begin():
            conn.execute(row_insert)

    def _get_all(self):
        rows_select = self.test_table.select()
        with self.engine.connect() as conn, conn.begin():
            return conn.execute(rows_select).fetchall()

    def _update_row(self, **kargs):
        row_update = self.test_table.update().values(**kargs)
        with self.engine.connect() as conn, conn.begin():
            conn.execute(row_update)

    def _delete_rows(self):
        row_delete = self.test_table.delete()
        with self.engine.connect() as conn, conn.begin():
            conn.execute(row_delete)


class IPAddressTestCase(SqlAlchemyTypesBaseTestCase):

    def _get_test_table(self, meta):
        return sa.Table(
            'fakeipaddressmodels',
            meta,
            sa.Column('id', sa.String(36), primary_key=True, nullable=False),
            sa.Column('ip', sqlalchemytypes.IPAddress))

    def _validate_ip_address(self, expected=None):
        objs = self._get_all()
        self.assertEqual(len(expected) if expected else 0, len(objs))
        if expected:
            for obj in objs:
                name = obj.id
                self.assertEqual(expected[name], obj.ip)

    def _test_crud(self, ip_addresses):
        ip = netaddr.IPAddress(ip_addresses[0])
        self._add_row(id='fake_id', ip=ip)
        self._validate_ip_address(expected={'fake_id': ip})

        ip2 = netaddr.IPAddress(ip_addresses[1])
        self._update_row(ip=ip2)
        self._validate_ip_address(expected={'fake_id': ip2})

        self._delete_rows()
        self._validate_ip_address(expected=None)

    def test_crud(self):
        ip_addresses = ["10.0.0.1", "10.0.0.2"]
        self._test_crud(ip_addresses)

        ip_addresses = ["2210::ffff:ffff:ffff:ffff",
                        "2120::ffff:ffff:ffff:ffff"]
        self._test_crud(ip_addresses)

    def test_wrong_type(self):
        self.assertRaises(exception.DBError, self._add_row,
                          id='fake_id', ip="")
        self.assertRaises(exception.DBError, self._add_row,
                          id='fake_id', ip="10.0.0.5")

    def _test_multiple_create(self, entries):
        reference = {}
        for entry in entries:
            ip = netaddr.IPAddress(entry['ip'])
            name = entry['name']
            self._add_row(id=name, ip=ip)
            reference[name] = ip

        self._validate_ip_address(expected=reference)
        self._delete_rows()
        self._validate_ip_address(expected=None)

    def test_multiple_create(self):
        ip_addresses = [
            {'name': 'fake_id1', 'ip': "10.0.0.5"},
            {'name': 'fake_id2', 'ip': "10.0.0.1"},
            {'name': 'fake_id3',
             'ip': "2210::ffff:ffff:ffff:ffff"},
            {'name': 'fake_id4',
             'ip': "2120::ffff:ffff:ffff:ffff"}]
        self._test_multiple_create(ip_addresses)


class CIDRTestCase(SqlAlchemyTypesBaseTestCase):

    def _get_test_table(self, meta):
        return sa.Table(
            'fakecidrmodels',
            meta,
            sa.Column('id', sa.String(36), primary_key=True, nullable=False),
            sa.Column('cidr', sqlalchemytypes.CIDR)
        )

    def _get_one(self, value):
        row_select = self.test_table.select().\
            where(self.test_table.c.cidr == value)
        with self.engine.connect() as conn, conn.begin():
            return conn.execute(row_select).first()

    def _update_row(self, key, cidr):
        row_update = self.test_table.update().values(cidr=cidr).\
                         where(self.test_table.c.cidr == key)
        with self.engine.connect() as conn, conn.begin():
            conn.execute(row_update)

    def test_crud(self):
        cidrs = ["10.0.0.0/24", "10.123.250.9/32", "2001:db8::/42",
                 "fe80::21e:67ff:fed0:56f0/64"]

        for cidr_str in cidrs:
            cidr = netaddr.IPNetwork(cidr_str)
            self._add_row(id=uuidutils.generate_uuid(), cidr=cidr)
            obj = self._get_one(cidr)
            self.assertEqual(cidr, obj.cidr)
            random_cidr = netaddr.IPNetwork(tools.get_random_cidr())
            self._update_row(cidr, random_cidr)
            obj = self._get_one(random_cidr)
            self.assertEqual(random_cidr, obj.cidr)

        objs = self._get_all()
        self.assertEqual(len(cidrs), len(objs))
        self._delete_rows()
        objs = self._get_all()
        self.assertEqual(0, len(objs))

    def test_wrong_cidr(self):
        wrong_cidrs = ["10.500.5.0/24", "10.0.0.1/40", "10.0.0.10.0/24",
                       "cidr", "", '2001:db8:5000::/64', '2001:db8::/130']
        for cidr in wrong_cidrs:
            self.assertRaises(exception.DBError, self._add_row,
                              id=uuidutils.generate_uuid(), cidr=cidr)


class MACAddressTestCase(SqlAlchemyTypesBaseTestCase):

    def _get_test_table(self, meta):
        return sa.Table(
            'fakemacaddressmodels',
            meta,
            sa.Column('id', sa.String(36), primary_key=True, nullable=False),
            sa.Column('mac', sqlalchemytypes.MACAddress)
        )

    def _get_one(self, value):
        row_select = self.test_table.select().\
            where(self.test_table.c.mac == value)
        with self.engine.connect() as conn, conn.begin():
            return conn.execute(row_select).first()

    def _get_all(self):
        rows_select = self.test_table.select()
        with self.engine.connect() as conn, conn.begin():
            return conn.execute(rows_select).fetchall()

    def _update_row(self, key, mac):
        row_update = self.test_table.update().values(mac=mac).\
                         where(self.test_table.c.mac == key)
        with self.engine.connect() as conn, conn.begin():
            conn.execute(row_update)

    def _delete_row(self):
        row_delete = self.test_table.delete()
        with self.engine.connect() as conn, conn.begin():
            conn.execute(row_delete)

    def test_crud(self):
        mac_addresses = ['FA:16:3E:00:00:01', 'FA:16:3E:00:00:02']

        for mac in mac_addresses:
            mac = netaddr.EUI(mac)
            self._add_row(id=uuidutils.generate_uuid(), mac=mac)
            obj = self._get_one(mac)
            self.assertEqual(mac, obj.mac)
            random_mac = netaddr.EUI(net.get_random_mac(
                ['fe', '16', '3e', '00', '00', '00']))
            self._update_row(mac, random_mac)
            obj = self._get_one(random_mac)
            self.assertEqual(random_mac, obj.mac)

        objs = self._get_all()
        self.assertEqual(len(mac_addresses), len(objs))
        self._delete_rows()
        objs = self._get_all()
        self.assertEqual(0, len(objs))

    def test_wrong_mac(self):
        wrong_macs = ["fake", "", -1,
                      "FK:16:3E:00:00:02",
                      "FA:16:3E:00:00:020"]
        for mac in wrong_macs:
            self.assertRaises(exception.DBError, self._add_row,
                              id=uuidutils.generate_uuid(), mac=mac)


class TruncatedDateTimeTestCase(SqlAlchemyTypesBaseTestCase):

    def _get_test_table(self, meta):
        return sa.Table(
            'timetable',
            meta,
            sa.Column('id', sa.String(36), primary_key=True, nullable=False),
            sa.Column('thetime', sqlalchemytypes.TruncatedDateTime)
        )

    def test_microseconds_truncated(self):
        tstamp = timeutils.utcnow()
        tstamp_low = tstamp.replace(microsecond=111111)
        tstamp_high = tstamp.replace(microsecond=999999)
        self._add_row(id=1, thetime=tstamp_low)
        self._add_row(id=2, thetime=tstamp_high)
        rows = self._get_all()
        self.assertEqual(2, len(rows))
        self.assertEqual(rows[0].thetime, rows[1].thetime)



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.db import constants as db_const


# The alias of the extension.
ALIAS = 'trunk'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'Trunk Extension'

# The description of the extension.
DESCRIPTION = "Provides support for trunk ports"

# A timestamp of when the extension was introduced.
UPDATED_TIMESTAMP = "2016-01-01T10:00:00-00:00"

# The specific resources and/or attributes for the extension (optional).
TRUNK = 'trunk'
TRUNKS = 'trunks'
SUB_PORTS = 'sub_ports'

# The resource attribute map for the extension.
RESOURCE_ATTRIBUTE_MAP = {
    TRUNKS: {
        'admin_state_up': {'allow_post': True, 'allow_put': True,
                           'default': True,
                           'convert_to': converters.convert_to_boolean,
                           'is_filter': True,
                           'is_sort_key': True,
                           'is_visible': True},
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_filter': True,
               'is_sort_key': True,
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': db_const.NAME_FIELD_SIZE},
                 'is_filter': True,
                 'is_sort_key': True,
                 'default': '', 'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'required_by_policy': True,
                      'validate':
                          {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                      'is_filter': True,
                      'is_sort_key': True,
                      'is_visible': True},
        'port_id': {'allow_post': True, 'allow_put': False,
                    'required_by_policy': True,
                    'validate': {'type:uuid': None},
                    'is_filter': True,
                    'is_sort_key': True,
                    'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_sort_key': True,
                   'is_visible': True},
        SUB_PORTS: {'allow_post': True, 'allow_put': False,
                    'default': [],
                    'convert_list_to': converters.convert_kvp_list_to_dict,
                    'validate': {'type:subports': None},
                    'enforce_policy': True,
                    'is_visible': True}
    },
}

# The subresource attribute map for the extension.
SUB_RESOURCE_ATTRIBUTE_MAP = None

# The action map.
ACTION_MAP = {
    TRUNK: {
        'add_subports': 'PUT',
        'remove_subports': 'PUT',
        'get_subports': 'GET'
    }
}

# The action status.
ACTION_STATUS = {
}

# The list of required extensions.
REQUIRED_EXTENSIONS = [
    "binding",
]

# The list of optional extensions.
OPTIONAL_EXTENSIONS = [
    "provider",  # needed to learn about network segmentation details.
]

# TODO(armax): add support for modeling custom queries



# Copyright (c) 2023 Ericsson Software Technology
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api.definitions import port


ALIAS = 'port-hints'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Port hints'
DESCRIPTION = 'Backend specific port hints to allow backend specific tuning'
UPDATED_TIMESTAMP = '2023-01-01T00:00:00-00:00'
RESOURCE_NAME = port.RESOURCE_NAME
COLLECTION_NAME = port.COLLECTION_NAME
HINTS = 'hints'
HINTS_SPEC = {
    'type:dict_or_none': {'openvswitch': {
        'type:dict': {'other_config': {
            'type:dict': {'tx-steering': {
                'type:values': ['thread', 'hash']}}}}}}}

RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        HINTS: {
            'allow_post': True,
            'allow_put': True,
            'default': None,
            'enforce_policy': True,
            'is_visible': True,
            'validate': HINTS_SPEC,
        },
    },
}

SUB_RESOURCE_ATTRIBUTE_MAP = None
ACTION_MAP = {}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from neutron_lib.api import converters
from neutron_lib.db import constants

ALIAS = 'agent'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = ALIAS
API_PREFIX = ''
DESCRIPTION = 'The agent management extension.'
UPDATED_TIMESTAMP = '2013-02-03T10:00:00-00:00'
RESOURCE_NAME = ALIAS
COLLECTION_NAME = ALIAS + 's'
RESOURCE_ATTRIBUTE_MAP: dict[str, typing.Any] = {
    COLLECTION_NAME: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_filter': True,
               'is_visible': True},
        'agent_type': {'allow_post': False, 'allow_put': False,
                       'is_filter': True, 'is_visible': True},
        'binary': {'allow_post': False, 'allow_put': False,
                   'is_filter': True, 'is_visible': True},
        'topic': {'allow_post': False, 'allow_put': False,
                  'is_filter': True, 'is_visible': True},
        'host': {'allow_post': False, 'allow_put': False,
                 'is_filter': True, 'is_visible': True},
        'admin_state_up': {'allow_post': False, 'allow_put': True,
                           'convert_to': converters.convert_to_boolean,
                           'is_filter': True,
                           'is_visible': True},
        'created_at': {'allow_post': False, 'allow_put': False,
                       'is_visible': True},
        'started_at': {'allow_post': False, 'allow_put': False,
                       'is_visible': True},
        'heartbeat_timestamp': {'allow_post': False, 'allow_put': False,
                                'is_visible': True},
        'alive': {'allow_post': False, 'allow_put': False,
                  'is_filter': True, 'is_visible': True},
        'configurations': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'description': {
            'allow_post': False, 'allow_put': True,
            'is_visible': True, 'is_filter': True,
            'validate': {
                'type:string_or_none': constants.DESCRIPTION_FIELD_SIZE}},
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.services import base
from neutron_lib.tests import _base as test_base


class _Worker(base.WorkerBase):
    pass


class Test_WorkerSupportServiceMixin(test_base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.worker = _Worker()

    def test_allocate_workers(self):
        self.assertEqual([], self.worker.get_workers())

    def test_add_worker(self):
        workers = [object(), object()]
        for w in workers:
            self.worker.add_worker(w)

        self.assertSequenceEqual(workers, self.worker.get_workers())

    def test_add_workers(self):
        workers = [object(), object(), object()]
        self.worker.add_workers(workers)

        self.assertSequenceEqual(workers, self.worker.get_workers())


class TestPluginInterface(test_base.BaseTestCase):

    class ServicePluginStub(base.ServicePluginBase):
        def get_plugin_type(self):
            pass

        def get_plugin_description(self):
            pass

    def test_issubclass_hook(self):
        class A(TestPluginInterface.ServicePluginStub):
            def f(self):
                pass

        class B(base.ServicePluginBase):
            @abc.abstractmethod
            def f(self):
                pass

        self.assertTrue(issubclass(A, B))

    def test_issubclass_hook_class_without_abstract_methods(self):
        class A:
            def f(self):
                pass

        class B(base.ServicePluginBase):
            def f(self):
                pass

        self.assertFalse(issubclass(A, B))

    def test_issubclass_hook_not_all_methods_implemented(self):
        class A(TestPluginInterface.ServicePluginStub):
            def f(self):
                pass

        class B(base.ServicePluginBase):
            @abc.abstractmethod
            def f(self):
                pass

            @abc.abstractmethod
            def g(self):
                pass

        self.assertFalse(issubclass(A, B))



# Copyright 2017 Eayun, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib.api.definitions import flavors
from neutron_lib.api.definitions import vpn

FLAVOR_ID = 'flavor_id'

ALIAS = 'vpn-flavors'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'VPN Service Flavor Extension'
DESCRIPTION = 'Flavor support for vpnservices.'
UPDATED_TIMESTAMP = '2017-04-19T00:00:00-00:00'
API_PREFIX = '/vpn'
RESOURCE_NAME = vpn.VPNSERVICE
COLLECTION_NAME = vpn.VPNSERVICES
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        FLAVOR_ID: {'allow_post': True, 'allow_put': False,
                    'validate': {'type:uuid_or_none': None},
                    'is_visible': True, 'default': None}
    },
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [vpn.ALIAS]
OPTIONAL_EXTENSIONS = [flavors.ALIAS]
ACTION_STATUS = {}



# Copyright (c) 2018 Ericsson
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api.definitions import agent
from neutron_lib import constants


ALIAS = 'agent-resources-synced'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = "Agent's Resource View Synced to Placement"
DESCRIPTION = 'Stores success/failure of last sync to Placement'
UPDATED_TIMESTAMP = '2018-12-19T00:00:00-00:00'
RESOURCE_NAME = agent.RESOURCE_NAME
COLLECTION_NAME = agent.COLLECTION_NAME
RESOURCES_SYNCED = 'resources_synced'

RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        RESOURCES_SYNCED: {
            'allow_post': False,
            'allow_put': False,
            'default': constants.ATTR_NOT_SPECIFIED,
            'is_visible': True,
        }
    }
}

SUB_RESOURCE_ATTRIBUTE_MAP = None
ACTION_MAP = {}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = [agent.ALIAS]
OPTIONAL_EXTENSIONS = []



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib._i18n import _
from neutron_lib import exceptions


class FirewallGroupNotFound(exceptions.NotFound):
    message = _("Firewall group %(firewall_id)s could not be found.")


class FirewallGroupInUse(exceptions.InUse):
    message = _("Firewall group %(firewall_id)s is still active.")


class FirewallGroupInPendingState(exceptions.Conflict):
    message = _("Operation cannot be performed since associated firewall "
                "group %(firewall_id)s is in %(pending_state)s.")


class FirewallGroupPortInvalid(exceptions.Conflict):
    message = _("Port %(port_id)s of firewall group is invalid.")


class FirewallGroupPortInvalidProject(exceptions.Conflict):
    message = _("Operation cannot be performed as port %(port_id)s "
                "is in an invalid project %(project_id)s.")


class FirewallGroupPortInUse(exceptions.InUse):
    message = _("Port(s) %(port_ids)s provided already associated with "
                "other firewall group(s).")


class FirewallPolicyNotFound(exceptions.NotFound):
    message = _("Firewall policy %(firewall_policy_id)s could not be found.")


class FirewallPolicyInUse(exceptions.InUse):
    message = _("Firewall policy %(firewall_policy_id)s is being used.")


class FirewallPolicyConflict(exceptions.NotFound):
    """FWaaS exception raised for firewall policy conflict

    Raised when user tries to use another project's unshared policy.
    """
    message = _("Operation cannot be performed since firewall policy "
                "%(firewall_policy_id)s for your project could not be found. "
                "Please confirm if the firewall policy exists and is shared.")


class FirewallRuleSharingConflict(exceptions.NotFound):
    """FWaaS exception for sharing policies

    Raised when shared policy uses unshared rules.
    """
    message = _("Operation cannot be performed since firewall policy "
                "%(firewall_policy_id)s could not find the firewall rule "
                "%(firewall_rule_id)s. Please confirm if the firewall rule "
                "exists and is shared.")


class FirewallPolicySharingConflict(exceptions.Conflict):
    """FWaaS exception raised for sharing policies

    Raised if policy is 'shared' but its associated rules are not.
    """
    message = _("Operation cannot be performed. Before sharing firewall "
                "policy %(firewall_policy_id)s, share associated firewall "
                "rule %(firewall_rule_id)s.")


class FirewallRuleNotFound(exceptions.NotFound):
    message = _("Firewall rule %(firewall_rule_id)s could not be found.")


class FirewallRuleInUse(exceptions.InUse):
    message = _("Firewall rule %(firewall_rule_id)s is being used.")


class FirewallRuleNotAssociatedWithPolicy(exceptions.InvalidInput):
    message = _("Firewall rule %(firewall_rule_id)s is not associated "
                "with firewall policy %(firewall_policy_id)s.")


class FirewallRuleInvalidProtocol(exceptions.InvalidInput):
    message = _("Protocol %(protocol)s is not supported. "
                "Only protocol values %(values)s and their integer "
                "representation (0 to 255) are supported.")


class FirewallRuleInvalidAction(exceptions.InvalidInput):
    message = _("Action %(action)s is not supported. "
                "Only action values %(values)s are supported.")


class FirewallRuleInvalidICMPParameter(exceptions.InvalidInput):
    message = _("%(param)s are not allowed when protocol "
                "is set to ICMP.")


class FirewallRuleWithPortWithoutProtocolInvalid(exceptions.InvalidInput):
    message = _("Source/destination port requires a protocol.")


class FirewallRuleInvalidPortValue(exceptions.InvalidInput):
    message = _("Invalid value for port %(port)s.")


class FirewallRuleInfoMissing(exceptions.InvalidInput):
    message = _("Missing rule info argument for insert/remove "
                "rule operation.")


class FirewallIpAddressConflict(exceptions.InvalidInput):
    message = _("Invalid input - IP addresses do not agree with IP Version.")


class FirewallInternalDriverError(exceptions.NeutronException):
    """FWaaS exception for all driver errors

    On any failure or exception in the driver, driver should log it and
    raise this exception to the agent
    """

    message = _("%(driver)s: Internal driver error.")


class FirewallRuleConflict(exceptions.Conflict):
    """FWaaS rule conflict exception

    Occurs when admin policy tries to use another project's rule that is
    not shared.
    """
    message = _("Operation cannot be performed since firewall rule "
                "%(firewall_rule_id)s is not shared and belongs to "
                "another project %(project_id)s.")


class FirewallRuleAlreadyAssociated(exceptions.Conflict):
    """FWaaS exception for an already associated rule

    Occurs when there is an attempt to assign a rule to a policy that
    the rule is already associated with.
    """
    message = _("Operation cannot be performed since firewall rule "
                "%(firewall_rule_id)s is already associated with firewall "
                "policy %(firewall_policy_id)s.")


class FirewallGroupCannotRemoveDefault(exceptions.InUse):
    message = _("Deleting default firewall group not allowed.")


class FirewallGroupCannotUpdateDefault(exceptions.InUse):
    message = _("Updating default firewall group not allowed.")


class FirewallGroupDefaultAlreadyExists(exceptions.InUse):
    """Default firewall group conflict exception

    Occurs when a user creates firewall group named 'default'.
    """
    message = _("Default firewall group already exists. 'default' is the "
                "reserved name for firewall group.")


class FirewallGroupPortNotSupported(exceptions.Conflict):
    message = _("Port %(port_id)s is not supported by firewall driver "
                "'%(driver_name)s'.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from neutron_lib._i18n import _
from neutron_lib.api.definitions import network
from neutron_lib.api.definitions import port
from neutron_lib.api.definitions import subnet
from neutron_lib.api.definitions import subnetpool
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib import exceptions


def _validate_privileges(context, res_dict):
    if ('project_id' in res_dict and
            res_dict['project_id'] != context.project_id and
            not (context.is_admin or context.is_advsvc)):
        msg = _("Specifying 'project_id' or 'tenant_id' other than the "
                "authenticated project in request requires admin or advsvc "
                "privileges")
        raise exc.HTTPBadRequest(msg)


def populate_project_info(attributes):
    """Ensure that both project_id and tenant_id attributes are present.

    If either project_id or tenant_id is present in attributes then ensure
    that both are present.

    If neither are present then attributes is not updated.

    :param attributes: A dictionary of resource/API attributes
        or API request/response dict.
    :returns: attributes (updated with project_id if applicable).
    :raises: HTTPBadRequest if the attributes project_id and tenant_id
        don't match.
    """
    if 'tenant_id' in attributes and 'project_id' not in attributes:
        attributes['project_id'] = attributes['tenant_id']
    elif 'project_id' in attributes and 'tenant_id' not in attributes:
        # Backward compatibility for code still using tenant_id
        attributes['tenant_id'] = attributes['project_id']

    if attributes.get('project_id') != attributes.get('tenant_id'):
        msg = _("'project_id' and 'tenant_id' do not match")
        raise exc.HTTPBadRequest(msg)

    return attributes


def _fill_default(res_dict, attr_name, attr_spec):
    # update res_dict[attr_name] record with the default value
    # specified in attr_spec, taking into account default_overrides_none
    if attr_spec.get('default_overrides_none'):
        if res_dict.get(attr_name) is None:
            res_dict[attr_name] = attr_spec.get('default')
            return

    res_dict[attr_name] = res_dict.get(attr_name,
                                       attr_spec.get('default'))


def _dict_populate_defaults(attr_value, attr_spec):
    # attr_value: dict
    # attr_spec: an attribute specification dict e.g.
    # {
    #      ...
    #      'dict_populate_defaults': True,
    #      'default_overrides_none': ,
    #      'validate': {
    #                 'type:dict': {
    #                     'foo': {
    #                         'default': FOO_DEFAULT,
    #                         'type:values': [42, 43]
    #                     },
    #                     'bar': {
    #                         'default': BAR_DEFAULT,
    #                         'convert_to': converters.convert_to_boolean
    #                     }
    #                     'baz': {
    #                         'dict_populate_defaults': True
    #                         'default_overrides_none': True,
    #                         'type:dict': {
    #                             'baz_bar: {
    #                                 'default': 77,
    #                                 'type:boolean'
    #                             },
    #                             'baz_foo': {
    #                                 'default': 88,
    #                                 'type:boolean'
    #                             }
    #                         }
    #                     }
    #             }
    #      }
    # }
    if not attr_spec.get(constants.DICT_POPULATE_DEFAULTS):
        return attr_value

    if attr_value is None or attr_value is constants.ATTR_NOT_SPECIFIED:
        attr_value = {}

    for rule_type, rule_content in attr_spec['validate'].items():
        # we only recursively apply defaults for dict rules
        if 'dict' not in rule_type:
            continue
        for key, key_validator in rule_content.items():
            validator_name, _dummy, validator_params = (
                validators._extract_validator(key_validator))

            # recurse if required:
            if 'dict' in validator_name:
                value = _dict_populate_defaults(
                    attr_value.get(key),
                    {
                        constants.DICT_POPULATE_DEFAULTS: key_validator.get(
                            constants.DICT_POPULATE_DEFAULTS),
                        'validate': {validator_name: validator_params}
                    }
                )
                if value is not None:
                    attr_value[key] = value

            _fill_default(attr_value, key, key_validator)

    return attr_value


class AttributeInfo:
    """Provides operations on a resource's attribute map.

    AttributeInfo wraps an API resource's attribute dict and provides methods
    for filling defaults, validating, converting, etc. based on the
    underlying attributes.
    """

    def __init__(self, resource_attrs):
        """Create a new instance that wraps the given resource attributes.

        :param resource_attrs: The resource's attributes that can be any
            of the following types: an instance of AttributeInfo, an API
            definition that contains a RESOURCE_ATTRIBUTE_MAP attribute or
            a dict of attributes for the resource.
        """
        if isinstance(resource_attrs, AttributeInfo):
            resource_attrs = resource_attrs.attributes
        elif getattr(resource_attrs,
                     'RESOURCE_ATTRIBUTE_MAP', None) is not None:
            # handle neutron_lib API definitions
            resource_attrs = resource_attrs.RESOURCE_ATTRIBUTE_MAP

        self.attributes = resource_attrs

    def fill_post_defaults(
            self, res_dict,
            exc_cls=lambda m: exceptions.InvalidInput(error_message=m),
            check_allow_post=True):
        """Fill in default values for attributes in a POST request.

        When a POST request is made, the attributes with default values do not
        need to be specified by the user. This function fills in the values of
        any unspecified attributes if they have a default value.

        If an attribute is not specified and it does not have a default value,
        an exception is raised.

        If an attribute is specified and it is not allowed in POST requests, an
        exception is raised. The caller can override this behavior by setting
        check_allow_post=False (used by some internal admin operations).

        :param res_dict: The resource attributes from the request.
        :param exc_cls: Exception to be raised on error that must take
            a single error message as it's only constructor arg.
        :param check_allow_post: Raises an exception if a non-POST-able
            attribute is specified.
        :raises: exc_cls If check_allow_post is True and this instance of
            ResourceAttributes doesn't support POST.
        """
        for attr, attr_vals in self.attributes.items():
            if attr_vals['allow_post']:
                value = _dict_populate_defaults(
                    res_dict.get(attr, constants.ATTR_NOT_SPECIFIED),
                    attr_vals)
                if value is not constants.ATTR_NOT_SPECIFIED:
                    res_dict[attr] = value

                if 'default' not in attr_vals and attr not in res_dict:
                    msg = _("Failed to parse request. Required "
                            "attribute '%s' not specified") % attr
                    raise exc_cls(msg)

                _fill_default(res_dict, attr, attr_vals)
            elif check_allow_post:
                if attr in res_dict:
                    msg = _("Attribute '%s' not allowed in POST") % attr
                    raise exc_cls(msg)

    def convert_values(
            self, res_dict,
            exc_cls=lambda m: exceptions.InvalidInput(error_message=m)):
        """Convert and validate attribute values for a request.

        :param res_dict: The resource attributes from the request.
        :param exc_cls: Exception to be raised on error that must take
            a single error message as it's only constructor arg.
        :raises: exc_cls If any errors occur converting/validating the
            res_dict.
        """
        for attr, attr_vals in self.attributes.items():
            if (attr not in res_dict or
                    res_dict[attr] is constants.ATTR_NOT_SPECIFIED):
                continue
            # Convert values if necessary
            if 'convert_to' in attr_vals:
                res_dict[attr] = attr_vals['convert_to'](res_dict[attr])
            # Check that configured values are correct
            if 'validate' not in attr_vals:
                continue
            for rule in attr_vals['validate']:
                validator = validators.get_validator(rule)
                res = validator(res_dict[attr],
                                attr_vals['validate'][rule])
                if res:
                    msg_dict = {'attr': attr, 'reason': res}
                    msg = _("Invalid input for %(attr)s. "
                            "Reason: %(reason)s.") % msg_dict
                    raise exc_cls(msg)

    def _project_id_required(self, res_dict):
        return (('tenant_id' in self.attributes or
                 'project_id' in self.attributes) and
                'project_id' not in res_dict)

    def populate_project_id(self, context, res_dict, is_create):
        """Populate the owner information in a request body.

        Ensure both project_id and tenant_id attributes are present.
        Validate that the requestor has the required privileges.
        For a create request, copy owner info from context to request body
        if needed and verify that owner is specified if required.

        :param context: The request context.
        :param res_dict: The resource attributes from the request.
        :param attr_info: The attribute map for the resource.
        :param is_create: Is this a create request?
        :raises: HTTPBadRequest If neither the project_id nor tenant_id
            are specified in the res_dict.

        """
        populate_project_info(res_dict)
        _validate_privileges(context, res_dict)

        if is_create and self._project_id_required(res_dict):
            if context.project_id:
                res_dict['project_id'] = context.project_id

                # For backward compatibility
                res_dict['tenant_id'] = context.project_id

            else:
                msg = _("Running without keystone AuthN requires "
                        "that tenant_id is specified")
                raise exc.HTTPBadRequest(msg)

    def verify_attributes(self, attrs_to_verify):
        """Reject unknown attributes.

        Consumers should ensure the project info is populated in the
        attrs_to_verify before calling this method.

        :param attrs_to_verify: The attributes to verify against this
            resource attributes.
        :raises: HTTPBadRequest: If attrs_to_verify contains any unrecognized
            for this resource attributes instance.
        """
        extra_keys = set(attrs_to_verify.keys()) - set(self.attributes.keys())
        if extra_keys:
            msg = _("Unrecognized attribute(s) '%s'") % ', '.join(extra_keys)
            raise exc.HTTPBadRequest(msg)


def _core_resource_attributes():
    resources = {}
    for core_def in [network.RESOURCE_ATTRIBUTE_MAP,
                     port.RESOURCE_ATTRIBUTE_MAP,
                     subnet.RESOURCE_ATTRIBUTE_MAP,
                     subnetpool.RESOURCE_ATTRIBUTE_MAP]:
        resources.update(core_def)
    return resources


# populate core resources into singleton global
RESOURCES = _core_resource_attributes()


def retrieve_valid_sort_keys(attr_info):
    """Retrieve sort keys from `attr_info` dict.

    Iterate the `attr_info`, filter and return the attributes that are
    defined with `is_sort_key=True`.

    :param attr_info: The attribute dict for common neutron resource.
    :returns: Set of sort keys.
    """
    return {attr for attr, schema in attr_info.items()
            if schema.get('is_sort_key', False)}



# Copyright (c) 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api.definitions import network as network_def
from neutron_lib.api import validators
from neutron_lib import constants


VLANTRANSPARENT = 'vlan_transparent'


def get_vlan_transparent(network):
    """Get the value of vlan_transparent from a network if set.

    :param network: The network dict to retrieve the value of vlan_transparent
        from.
    :returns: The value of vlan_transparent from the network dict if set in
        the dict, otherwise False is returned.
    """
    return (network[VLANTRANSPARENT]
            if (VLANTRANSPARENT in network and
                validators.is_attr_set(network[VLANTRANSPARENT]))
            else False)


ALIAS = 'vlan-transparent'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Vlantransparent'
API_PREFIX = ''
DESCRIPTION = 'Provides Vlan Transparent Networks'
UPDATED_TIMESTAMP = '2015-03-23T09:00:00-00:00'
RESOURCE_ATTRIBUTE_MAP = {
    network_def.COLLECTION_NAME: {
        VLANTRANSPARENT: {
            'allow_post': True,
            'allow_put': False,
            'convert_to': converters.convert_to_boolean,
            'default': constants.ATTR_NOT_SPECIFIED,
            'is_visible': True,
            'is_filter': True
        }
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins.ml2 import api
from neutron_lib.tests import _base as base


class _MechanismDriver(api.MechanismDriver):
    def bind_port(s, c):
        return c

    def initialize(self):
        pass


class TestMechanismDriver(base.BaseTestCase):

    def test__supports_port_binding(self):
        self.assertTrue(_MechanismDriver()._supports_port_binding)

    def test_get_workers(self):
        self.assertEqual((), _MechanismDriver().get_workers())

    def test_filter_hosts_with_segment_access(self):
        dummy_token = ["X"]
        self.assertEqual(
            dummy_token,
            _MechanismDriver().filter_hosts_with_segment_access(
                dummy_token, dummy_token, dummy_token, dummy_token))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from neutron_lib import context
from neutron_lib.db import model_base

from neutron_lib.tests.unit.db import _base as db_base


class TestTable(model_base.BASEV2, model_base.HasProject,
                model_base.HasId, model_base.HasStatusDescription):

    name = sa.Column(sa.String(8), primary_key=True)


class TestModelBase(db_base.SqlTestCase):

    def setUp(self):
        super().setUp()
        self.ctx = context.Context('user', 'project')
        self.session = self.ctx.session

    def test_model_base(self):
        foo = TestTable(name='meh')
        self.assertEqual('meh', foo.name)
        self.assertIn('meh', str(foo))  # test foo.__repr__
        cols = [k for k, _v in foo]  # test foo.__iter__ and foo.next
        self.assertIn('name', cols)

    def test_get_set_tenant_id_tenant(self):
        foo = TestTable(tenant_id='tenant')
        self.assertEqual('tenant', foo.get_tenant_id())
        foo.set_tenant_id('project')
        self.assertEqual('project', foo.get_tenant_id())

    def test_get_set_tenant_id_project(self):
        foo = TestTable(project_id='project')
        self.assertEqual('project', foo.get_tenant_id())
        foo.set_tenant_id('tenant')
        self.assertEqual('tenant', foo.get_tenant_id())

    def test_project_id_attribute(self):
        foo = TestTable(project_id='project')
        self.assertEqual('project', foo.project_id)
        self.assertEqual('project', foo.tenant_id)

    def test_tenant_id_attribute(self):
        foo = TestTable(tenant_id='tenant')
        self.assertEqual('tenant', foo.project_id)
        self.assertEqual('tenant', foo.tenant_id)



# Copyright (c) 2015 Huawei Technologies Co.,LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib import constants
from neutron_lib.db import constants as db_constants


ADDRESS_SCOPE = 'address_scope'
ADDRESS_SCOPE_ID = 'address_scope_id'
IPV4_ADDRESS_SCOPE = 'ipv4_%s' % ADDRESS_SCOPE
IPV6_ADDRESS_SCOPE = 'ipv6_%s' % ADDRESS_SCOPE

ALIAS = 'address-scope'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Address scope'
API_PREFIX = ''
DESCRIPTION = 'Address scopes extension.'
UPDATED_TIMESTAMP = '2015-07-26T10:00:00-00:00'
RESOURCE_NAME = ADDRESS_SCOPE
COLLECTION_NAME = RESOURCE_NAME + 's'
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        'id': {'allow_post': False,
               'allow_put': False,
               'validate': {'type:uuid': None},
               'is_filter': True,
               'is_sort_key': True,
               'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True,
                 'allow_put': True,
                 'default': '',
                 'validate': {'type:string': db_constants.NAME_FIELD_SIZE},
                 'is_filter': True,
                 'is_sort_key': True,
                 'is_visible': True},
        'tenant_id': {'allow_post': True,
                      'allow_put': False,
                      'validate': {
                          'type:string': db_constants.PROJECT_ID_FIELD_SIZE},
                      'required_by_policy': True,
                      'is_filter': True,
                      'is_sort_key': True,
                      'is_visible': True},
        constants.SHARED: {'allow_post': True,
                           'allow_put': True,
                           'default': False,
                           'convert_to': converters.convert_to_boolean,
                           'is_visible': True,
                           'is_filter': True,
                           'is_sort_key': True,
                           'required_by_policy': True,
                           'enforce_policy': True},
        'ip_version': {'allow_post': True, 'allow_put': False,
                       'convert_to': converters.convert_to_int,
                       'validate': {'type:values': [4, 6]},
                       'is_filter': True,
                       'is_sort_key': True,
                       'is_visible': True},
    },
    'subnetpools': {
        ADDRESS_SCOPE_ID: {'allow_post': True,
                           'allow_put': True,
                           'default': constants.ATTR_NOT_SPECIFIED,
                           'validate': {'type:uuid_or_none': None},
                           'is_filter': True,
                           'is_sort_key': True,
                           'is_visible': True}
    },
    'networks': {
        IPV4_ADDRESS_SCOPE: {'allow_post': False,
                             'allow_put': False,
                             'is_visible': True},
        IPV6_ADDRESS_SCOPE: {'allow_post': False,
                             'allow_put': False,
                             'is_visible': True},
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import network


MTU = 'mtu'

# The alias of the extension.
ALIAS = 'net-mtu'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'Network MTU'

# A prefix for API resources. An empty prefix means that the API is going
# to be exposed at the v2/ level as any other core resource.
API_PREFIX = ''

# The description of the extension.
DESCRIPTION = "Provides MTU attribute for a network resource."

# A timestamp of when the extension was introduced.
UPDATED_TIMESTAMP = "2015-03-25T10:00:00-00:00"


RESOURCE_ATTRIBUTE_MAP = {
    network.COLLECTION_NAME: {
        MTU: {'allow_post': False, 'allow_put': False,
              'is_visible': True, 'is_filter': True, 'is_sort_key': True},
    },
}

# The subresource attribute map for the extension. It adds child resources
# to main extension's resource. The subresource map must have a parent and
# a parameters entry. If an extension does not need such a map, None can
# be specified (mandatory).
SUB_RESOURCE_ATTRIBUTE_MAP = {}

# The action map: it associates verbs with methods to be performed on
# the API resource.
ACTION_MAP = {}

# The action status.
ACTION_STATUS = {
}

# The list of required extensions.
REQUIRED_EXTENSIONS = []

# The list of optional extensions.
OPTIONAL_EXTENSIONS = []



# Copyright (c) 2020 Troila.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import l3
from neutron_lib.db import constants as db_const

# The alias of the extension.
ALIAS = 'l3-ndp-proxy'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'L3 NDP Proxy'

# A prefix for API resources. An empty prefix means that the API is going
# to be exposed at the v2/ level as any other core resource.
API_PREFIX = ''

# The description of the extension.
DESCRIPTION = "Router support for L3 NDP proxy."

# A timestamp of when the extension was introduced.
UPDATED_TIMESTAMP = "2021-06-29T10:00:00-00:00"

# The name of the resource.
RESOURCE_NAME = 'ndp_proxy'

# The plural for the resource.
COLLECTION_NAME = 'ndp_proxies'

ID = 'id'
NAME = 'name'
PROJECT_ID = 'project_id'
ROUTER_ID = 'router_id'
PORT_ID = 'port_id'
IP_ADDRESS = 'ip_address'
DESCRIPTION = 'description'

# The resource attribute map for the extension.
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        ID: {'allow_post': False,
             'allow_put': False,
             'validate': {'type:uuid': None},
             'is_visible': True,
             'is_sort_key': True,
             'primary_key': True},
        NAME: {'allow_post': True,
               'allow_put': True,
               'validate': {'type:string': db_const.NAME_FIELD_SIZE},
               'is_filter': True,
               'is_sort_key': True,
               'is_visible': True, 'default': ''},
        PROJECT_ID: {'allow_post': True,
                     'allow_put': False,
                     'required_by_policy': True,
                     'validate': {
                         'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                     'is_sort_key': True,
                     'is_visible': True},
        ROUTER_ID: {'allow_post': True,
                    'allow_put': False,
                    'validate': {'type:uuid': None},
                    'is_sort_key': True,
                    'is_visible': True},
        PORT_ID: {'allow_post': True,
                  'allow_put': False,
                  'validate': {'type:uuid': None},
                  'is_sort_key': True,
                  'is_visible': True},
        IP_ADDRESS: {'allow_post': True,
                     'allow_put': False,
                     'default': None,
                     'validate': {
                         'type:ip_address_or_none': None},
                     'is_sort_key': True,
                     'is_visible': True},
        DESCRIPTION: {'allow_post': True,
                      'allow_put': True,
                      'default': '',
                      'validate': {'type:string':
                                   db_const.LONG_DESCRIPTION_FIELD_SIZE},
                      'is_visible': True}
    }
}

# The subresource attribute map for the extension. It adds child resources
# to main extension's resource. The subresource map must have a parent and
# a parameters entry. If an extension does not need such a map, None can
# be specified (mandatory).
SUB_RESOURCE_ATTRIBUTE_MAP = {}

# The action map: it associates verbs with methods to be performed on
# the API resource.
ACTION_MAP = {
}

# The action status.
ACTION_STATUS = {
}

# The list of required extensions.
REQUIRED_EXTENSIONS = [l3.ALIAS]

# The list of optional extensions.
OPTIONAL_EXTENSIONS = []



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import typing

from neutron_lib.api.definitions import l3
from neutron_lib.api.definitions import l3_ext_gw_mode
from neutron_lib.api.definitions import qos
from neutron_lib.services.qos import constants as qos_consts


ALIAS = 'qos-gateway-ip'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Router gateway IP QoS'
API_PREFIX = ''
DESCRIPTION = 'The Router gateway IP Quality of Service extension'
UPDATED_TIMESTAMP = '2018-02-24T00:00:00-00:00'
RESOURCE_NAME = l3.ROUTER
COLLECTION_NAME = l3.ROUTERS

routers: typing.Dict[str, typing.Any] = copy.deepcopy(
    l3_ext_gw_mode.RESOURCE_ATTRIBUTE_MAP[COLLECTION_NAME]
)
routers[l3.EXTERNAL_GW_INFO]['validate']['type:dict_or_nodata'][
    qos_consts.QOS_POLICY_ID] = {
    'type:uuid_or_none': None,
    'required': False
}
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: routers
}

SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [l3.ALIAS, qos.ALIAS, l3_ext_gw_mode.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
import testtools

from neutron_lib.api import converters
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from neutron_lib.tests import _base as base
from neutron_lib.tests import tools


class TestConvertToBoolean(base.BaseTestCase):

    def test_convert_to_boolean_bool(self):
        self.assertIs(converters.convert_to_boolean(True), True)
        self.assertIs(converters.convert_to_boolean(False), False)

    def test_convert_to_boolean_int(self):
        self.assertIs(converters.convert_to_boolean(0), False)
        self.assertIs(converters.convert_to_boolean(1), True)
        self.assertRaises(n_exc.InvalidInput,
                          converters.convert_to_boolean,
                          7)

    def test_convert_to_boolean_str(self):
        self.assertIs(converters.convert_to_boolean('True'), True)
        self.assertIs(converters.convert_to_boolean('true'), True)
        self.assertIs(converters.convert_to_boolean('False'), False)
        self.assertIs(converters.convert_to_boolean('false'), False)
        self.assertIs(converters.convert_to_boolean('0'), False)
        self.assertIs(converters.convert_to_boolean('1'), True)
        self.assertRaises(n_exc.InvalidInput,
                          converters.convert_to_boolean,
                          '7')

    def test_convert_to_boolean_if_not_none(self):
        self.assertIsNone(converters.convert_to_boolean_if_not_none(None))
        self.assertIs(converters.convert_to_boolean_if_not_none(1), True)


class TestConvertToInt(base.BaseTestCase):

    def test_convert_to_int_int(self):
        self.assertEqual(-1, converters.convert_to_int(-1))
        self.assertEqual(0, converters.convert_to_int(0))
        self.assertEqual(1, converters.convert_to_int(1))

    def test_convert_to_int_if_not_none(self):
        self.assertEqual(-1, converters.convert_to_int_if_not_none(-1))
        self.assertEqual(0, converters.convert_to_int_if_not_none(0))
        self.assertEqual(1, converters.convert_to_int_if_not_none(1))
        self.assertIsNone(converters.convert_to_int_if_not_none(None))

    def test_convert_to_int_str(self):
        self.assertEqual(4, converters.convert_to_int('4'))
        self.assertEqual(6, converters.convert_to_int('6'))
        self.assertRaises(n_exc.InvalidInput,
                          converters.convert_to_int,
                          'garbage')

    def test_convert_to_int_none(self):
        self.assertRaises(n_exc.InvalidInput,
                          converters.convert_to_int,
                          None)

    def test_convert_none_to_empty_list_none(self):
        self.assertEqual([], converters.convert_none_to_empty_list(None))

    def test_convert_none_to_empty_dict(self):
        self.assertEqual({}, converters.convert_none_to_empty_dict(None))

    def test_convert_none_to_empty_list_value(self):
        values = ['1', 3, [], [1], {}, {'a': 3}]
        for value in values:
            self.assertEqual(
                value, converters.convert_none_to_empty_list(value))


class TestConvertToFloat(base.BaseTestCase):
    # NOTE: the routine being tested here is a plugin-specific extension
    # module. As the plugin split proceed towards its second phase this
    # test should either be remove, or the validation routine moved into
    # neutron.api.v2.attributes

    def test_convert_to_float_positve_value(self):
        self.assertEqual(
            1.111, converters.convert_to_positive_float_or_none(1.111))
        self.assertEqual(1, converters.convert_to_positive_float_or_none(1))
        self.assertEqual(0, converters.convert_to_positive_float_or_none(0))

    def test_convert_to_float_negative_value(self):
        self.assertRaises(n_exc.InvalidInput,
                          converters.convert_to_positive_float_or_none,
                          -1.11)

    def test_convert_to_float_string(self):
        self.assertEqual(4, converters.convert_to_positive_float_or_none('4'))
        self.assertEqual(
            4.44, converters.convert_to_positive_float_or_none('4.44'))
        self.assertRaises(n_exc.InvalidInput,
                          converters.convert_to_positive_float_or_none,
                          'garbage')

    def test_convert_to_float_none_value(self):
        self.assertIsNone(converters.convert_to_positive_float_or_none(None))


class TestConvertKvp(base.BaseTestCase):

    def test_convert_kvp_list_to_dict_succeeds_for_missing_values(self):
        result = converters.convert_kvp_list_to_dict(['True'])
        self.assertEqual({}, result)

    def test_convert_kvp_list_to_dict_succeeds_for_multiple_values(self):
        result = converters.convert_kvp_list_to_dict(
            ['a=b', 'a=c', 'a=c', 'b=a'])
        expected = {'a': tools.UnorderedList(['c', 'b']), 'b': ['a']}
        self.assertEqual(expected, result)

    def test_convert_kvp_list_to_dict_succeeds_for_values(self):
        result = converters.convert_kvp_list_to_dict(['a=b', 'c=d'])
        self.assertEqual({'a': ['b'], 'c': ['d']}, result)

    def test_convert_kvp_str_to_list_fails_for_missing_key(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            converters.convert_kvp_str_to_list('=a')

    def test_convert_kvp_str_to_list_fails_for_missing_equals(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            converters.convert_kvp_str_to_list('a')

    def test_convert_kvp_str_to_list_succeeds_for_one_equals(self):
        result = converters.convert_kvp_str_to_list('a=')
        self.assertEqual(['a', ''], result)

    def test_convert_kvp_str_to_list_succeeds_for_two_equals(self):
        result = converters.convert_kvp_str_to_list('a=a=a')
        self.assertEqual(['a', 'a=a'], result)


class TestConvertToList(base.BaseTestCase):

    def test_convert_to_empty_list(self):
        for item in (None, [], (), {}):
            self.assertEqual([], converters.convert_to_list(item))

    def test_convert_to_list_string(self):
        for item in ('', 'foo'):
            self.assertEqual([item], converters.convert_to_list(item))

    def test_convert_to_list_iterable(self):
        for item in ([None], [1, 2, 3], (1, 2, 3), {1, 2, 3}, ['foo']):
            self.assertEqual(list(item), converters.convert_to_list(item))

    def test_convert_to_list_non_iterable(self):
        for item in (True, False, 1, 1.2, object()):
            self.assertEqual([item], converters.convert_to_list(item))


class TestConvertIPv6AddrCanonicalFormat(base.BaseTestCase):

    def test_convert_ipv6_address_extended_add_with_zeroes(self):
        result = converters.convert_ip_to_canonical_format(
            '2001:0db8:0:0:0:0:0:0001')
        self.assertEqual('2001:db8::1', result)

    def test_convert_ipv6_compressed_address(self):
        result = converters.convert_ip_to_canonical_format(
            '2001:db8:0:1:1:1:1:1')
        self.assertEqual('2001:db8:0:1:1:1:1:1', result)

    def test_convert_ipv6_extended_addr_to_compressed(self):
        result = converters.convert_ip_to_canonical_format(
            "Fe80:0:0:0:0:0:0:1")
        self.assertEqual('fe80::1', result)

    def test_convert_ipv4_address(self):
        result = converters.convert_ip_to_canonical_format("192.168.1.1")
        self.assertEqual('192.168.1.1', result)

    def test_convert_None_address(self):
        result = converters.convert_ip_to_canonical_format(None)
        self.assertIsNone(result)

    def test_convert_invalid_address(self):
        result = converters.convert_ip_to_canonical_format("on")
        self.assertEqual("on", result)
        result = converters.convert_ip_to_canonical_format(
            '192.168.1.1/32')
        self.assertEqual('192.168.1.1/32', result)
        result = converters.convert_ip_to_canonical_format(
            '2001:db8:0:1:1:1:1:1/128')
        self.assertEqual('2001:db8:0:1:1:1:1:1/128', result)

    def test_convert_subnetpools(self):
        pools = [{'start': '1.1.1.1', 'end': '1.1.1.100'}]
        result = converters.convert_ip_to_canonical_format(pools)
        self.assertEqual(pools, result)


class TestConvertAllocationPoolsCanonicalFormat(base.BaseTestCase):

    def test_convert_allocation_pools_to_canonical_format_noop(self):
        pools = [{'start': '1.1.1.1', 'end': '1.1.1.100'}]
        result = converters.convert_allocation_pools_to_canonical_format(pools)
        self.assertEqual(pools, result)

    def test_convert_allocation_pools_to_canonical_format(self):
        pools = [
            {'start': 'Fe80:0:0:0:0:0:0:1', 'end': 'Fe80:0:0:0:0:0:0:100'},
        ]
        result = converters.convert_allocation_pools_to_canonical_format(pools)

        expected = [
            {'start': 'fe80::1', 'end': 'fe80::100'},
        ]
        self.assertEqual(expected, result)

    def test_convert_allocation_pools_to_canonical_format_None(self):
        result = converters.convert_allocation_pools_to_canonical_format(None)
        self.assertEqual([], result)

    def test_convert_allocation_pools_to_canonical_format_invalid_data(self):
        self.assertRaises(
            n_exc.InvalidInput,
            converters.convert_allocation_pools_to_canonical_format,
            '10.0.0.0/24')


class TestConvertIPv6CIDRCanonicalFormat(base.BaseTestCase):

    def test_convert_ipv4_address_with_CIDR(self):
        result = converters.convert_cidr_to_canonical_format('192.168.1.1/24')
        self.assertEqual('192.168.1.1/24', result)

    def test_convert_ipv6_extended_addr_withcidr_to_compressed(self):
        result = converters.convert_cidr_to_canonical_format(
            'Fe80:0:0:0:0:0:0:1/64')
        self.assertEqual('fe80::1/64', result)

    def test_convert_non_ip_addr_with_slash(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            converters.convert_cidr_to_canonical_format(
                "Dormamu/DarkSeid/Vulture")


class TestConvertStringToCaseInsensitive(base.BaseTestCase):

    def test_convert_string_to_lower(self):
        result = converters.convert_string_to_case_insensitive("THIS Is tEsT")
        self.assertIsInstance(result, str)

    def test_assert_error_on_non_string(self):
        for invalid in [[], 123]:
            with testtools.ExpectedException(n_exc.InvalidInput):
                converters.convert_string_to_case_insensitive(invalid)


class TestConvertProtocol(base.BaseTestCase):

    def test_tcp_is_valid(self):
        result = converters.convert_to_protocol(constants.PROTO_NAME_TCP)
        self.assertEqual(constants.PROTO_NAME_TCP, result)
        proto_num_str = str(constants.PROTO_NUM_TCP)
        result = converters.convert_to_protocol(proto_num_str)
        self.assertEqual(proto_num_str, result)

    def test_udp_is_valid(self):
        result = converters.convert_to_protocol(constants.PROTO_NAME_UDP)
        self.assertEqual(constants.PROTO_NAME_UDP, result)
        proto_num_str = str(constants.PROTO_NUM_UDP)
        result = converters.convert_to_protocol(proto_num_str)
        self.assertEqual(proto_num_str, result)

    def test_icmp_is_valid(self):
        result = converters.convert_to_protocol(constants.PROTO_NAME_ICMP)
        self.assertEqual(constants.PROTO_NAME_ICMP, result)
        proto_num_str = str(constants.PROTO_NUM_ICMP)
        result = converters.convert_to_protocol(proto_num_str)
        self.assertEqual(proto_num_str, result)

    def test_numeric_is_valid(self):
        proto_num_str = str(constants.PROTO_NUM_IGMP)
        result = converters.convert_to_protocol(proto_num_str)
        self.assertEqual(proto_num_str, result)

    def test_numeric_too_high(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            converters.convert_to_protocol("300")

    def test_numeric_too_low(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            converters.convert_to_protocol("-1")

    def test_unknown_string(self):
        with testtools.ExpectedException(n_exc.InvalidInput):
            converters.convert_to_protocol("Invalid")


class TestConvertToString(base.BaseTestCase):

    def test_data_is_string(self):
        self.assertEqual('10000', converters.convert_to_string('10000'))

    def test_data_is_integer(self):
        self.assertEqual('10000', converters.convert_to_string(10000))

    def test_data_is_integer_zero(self):
        self.assertEqual('0', converters.convert_to_string(0))

    def test_data_is_none(self):
        self.assertIsNone(converters.convert_to_string(None))

    def test_data_is_empty_list(self):
        self.assertEqual('[]', converters.convert_to_string([]))

    def test_data_is_list(self):
        self.assertEqual("[1, 2, 3]", converters.convert_to_string([1, 2, 3]))

    def test_data_is_empty_dict(self):
        self.assertEqual('{}', converters.convert_to_string({}))

    def test_data_is_dict(self):
        self.assertEqual("{'foo': 'bar'}",
                         converters.convert_to_string({'foo': 'bar'}))


class TestConvertUppercasePrefix(base.BaseTestCase):

    def test_prefix_not_present(self):
        self.assertEqual('foobar',
                         converters.convert_prefix_forced_case('foobar',
                                                               'bar'))

    def test_prefix_no_need_to_replace(self):
        self.assertEqual('FOObar',
                         converters.convert_prefix_forced_case('FOObar',
                                                               'FOO'))

    def test_ucfirst_prefix_converted_1(self):
        self.assertEqual('Foobar',
                         converters.convert_prefix_forced_case('foobar',
                                                               'Foo'))

    def test_lc_prefix_converted_2(self):
        self.assertEqual('foobar',
                         converters.convert_prefix_forced_case('fOobar',
                                                               'foo'))

    def test_mixed_prefix_converted_1(self):
        self.assertEqual('fOoXbar',
                         converters.convert_prefix_forced_case('Fooxbar',
                                                               'fOoX'))

    def test_shorter_string(self):
        self.assertEqual('fo',
                         converters.convert_prefix_forced_case('fo',
                                                               'foo'))


class TestConvertPortMacAddress(base.BaseTestCase):

    def test_mac_address_does_not_convert(self):
        valid_mac = 'fa:16:3e:b6:78:1f'
        self.assertEqual(valid_mac,
                         converters.convert_to_mac_if_none(valid_mac))

    @mock.patch('oslo_config.cfg.CONF')
    def test_convert_none_to_mac_address(self, CONF):
        CONF.base_mac = 'fa:16:3e:00:00:00'
        self.assertTrue(
            netaddr.valid_mac(converters.convert_to_mac_if_none(None)))


class TestConvertToSanitizedMacAddress(base.BaseTestCase):

    def test_sanitize_mac_address(self):
        input_exp = (('00:11:22:33:44:55', '00:11:22:33:44:55'),
                     ('00:11:22:33:44:5', '00:11:22:33:44:05'),
                     ('0:1:2:3:4:5', '00:01:02:03:04:05'),
                     ('ca:FE:cA:Fe:a:E', 'ca:fe:ca:fe:0a:0e'),
                     ('12345678901', '01:23:45:67:89:01'),
                     ('012345678901', '01:23:45:67:89:01'),
                     )
        for input, expected in input_exp:
            self.assertEqual(
                expected,
                converters.convert_to_sanitized_mac_address(input))
            eui_address = netaddr.EUI(input)
            self.assertEqual(
                expected,
                converters.convert_to_sanitized_mac_address(eui_address))

        # Not converted, those MAC address will fail in the validation step.
        self.assertEqual(
            '00:11:22:33:44',
            converters.convert_to_sanitized_mac_address('00:11:22:33:44'))
        self.assertEqual(
            '00:11:22:33:44:',
            converters.convert_to_sanitized_mac_address('00:11:22:33:44:'))


class TestConvertToSanitizedBindingProfileAllocation(base.BaseTestCase):

    RP_UUID = '41d7391e-1f69-11ec-a899-8f9d6d950f8d'
    PORT_ID = '64d01804-1f83-11ec-987c-7f6caec3998b'
    MIN_BW_RULE_ID = '52441596-1f83-11ec-93c5-9b759591a493'
    # uuid -v5 64d01804-1f83-11ec-987c-7f6caec3998b
    # 52441596-1f83-11ec-93c5-9b759591a493
    GROUP_UUID = '2a1be6ea-15b0-5ac1-9d70-643e2ae306cb'

    def test_sanitize_binding_profile_allocation(self):
        old_format = self.RP_UUID
        new_format = {self.GROUP_UUID: self.RP_UUID}
        min_bw_rules = [mock.MagicMock(id=self.MIN_BW_RULE_ID)]
        self.assertEqual(
            new_format,
            converters.convert_to_sanitized_binding_profile_allocation(
                old_format, self.PORT_ID, min_bw_rules))

        self.assertEqual(
            new_format,
            converters.convert_to_sanitized_binding_profile_allocation(
                new_format, self.PORT_ID, min_bw_rules))



# Copyright (c) 2013 OpenStack Foundation.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import agent as agent_apidef
from neutron_lib.api.definitions import network as net_apidef
from neutron_lib import constants


DHCP_NET = 'dhcp-network'
DHCP_NETS = DHCP_NET + 's'
DHCP_AGENT = 'dhcp-agent'
DHCP_AGENTS = DHCP_AGENT + 's'

ALIAS = constants.DHCP_AGENT_SCHEDULER_EXT_ALIAS
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'DHCP Agent Scheduler'
API_PREFIX = ''
DESCRIPTION = 'Schedule networks among dhcp agents'
UPDATED_TIMESTAMP = '2013-02-07T10:00:00-00:00'
RESOURCE_ATTRIBUTE_MAP = {}
SUB_RESOURCE_ATTRIBUTE_MAP = {
    DHCP_NETS: {
        'parent': {
            'collection_name': agent_apidef.COLLECTION_NAME,
            'member_name': agent_apidef.RESOURCE_NAME
        },
        'parameters': {
            'network_id': {
                'allow_post': True, 'allow_put': False,
                'default': constants.ATTR_NOT_SPECIFIED,
                'enforce_policy': True,
                'is_visible': True,
                'validate': {'type:uuid': None}
            }
        }
    },
    DHCP_AGENTS: {
        'parent': {
            'collection_name': net_apidef.COLLECTION_NAME,
            'member_name': net_apidef.RESOURCE_NAME
        },
        # NOTE(boden): the reference implementation only allows the index
        # operation for the agent exposed under the network resource
        'parameters': agent_apidef.RESOURCE_ATTRIBUTE_MAP[
            agent_apidef.COLLECTION_NAME]
    }
}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [agent_apidef.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class AgentExtension(metaclass=abc.ABCMeta):
    """Define stable abstract interface for agent extensions.

    An agent extension extends the agent core functionality.
    """

    @abc.abstractmethod
    def initialize(self, connection, driver_type):
        """Perform agent core resource extension initialization.

        :param connection: RPC connection that can be reused by the extension
                           to define its RPC endpoints
        :param driver_type: String that defines the agent type to the
                            extension. Can be used to choose the right backend
                            implementation.

        Called after all extensions have been loaded.
        No resource (port, policy, router, etc.) handling will be called before
        this method.
        """

    def consume_api(self, agent_api):
        """Consume the AgentAPI instance from the AgentExtensionsManager.

        Allows an extension to gain access to resources internal to the
        neutron agent and otherwise unavailable to the extension.  Examples of
        such resources include bridges, ports, and routers.

        :param agent_api: An instance of an agent-specific API.
        """



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from neutron_lib._i18n import _
from neutron_lib import exceptions

LOG = logging.getLogger(__name__)


class MaxVRIDAllocationTriesReached(exceptions.NeutronException):
    message = _("Failed to allocate a VRID in the network %(network_id)s "
                "for the router %(router_id)s after %(max_tries)s tries.")


class NoVRIDAvailable(exceptions.Conflict):
    message = _("No more Virtual Router Identifier (VRID) available when "
                "creating router %(router_id)s. The limit of number "
                "of HA Routers per tenant is 254.")


class HANetworkConcurrentDeletion(exceptions.Conflict):
    message = _("Network for project %(project_id)s concurrently deleted.")

    # NOTE(haleyb): remove fall-back and warning in E+2 release, or when
    # all callers have been changed to use project_id.
    def __init__(self, **kwargs):
        project_id = kwargs.get('project_id')
        tenant_id = kwargs.get('tenant_id')
        project_id = project_id or tenant_id
        if tenant_id:
            LOG.warning('Keyword tenant_id has been deprecated, use '
                        'project_id instead')
        kwargs.setdefault('project_id', project_id)
        super().__init__(**kwargs)


class HANetworkCIDRNotValid(exceptions.NeutronException):
    message = _("The HA Network CIDR specified in the configuration file "
                "isn't valid; %(cidr)s.")


class HAMaximumAgentsNumberNotValid(exceptions.NeutronException):
    message = _("max_l3_agents_per_router %(max_agents)s config parameter "
                "is not valid as it cannot be negative. It must be 1 or "
                "greater. Alternatively, it can be 0 to mean unlimited.")


class DuplicatedHANetwork(exceptions.Conflict):
    message = _('Project %(project_id)s already has a HA network.')
